#!/usr/bin/env python3

import trojanzoo.optim

from trojanzoo.utils import add_noise, cos_sim
from trojanzoo.utils.output import prints

import torch
import torch.autograd
from collections.abc import Callable
from typing import Union


class PGD(trojanzoo.optim.Optimizer):
    r"""Projected Gradient Descent.
    Args:
        alpha (float): learning rate :math:`\alpha`. Default: :math:`\frac{3}{255}`.
        epsilon (float): the perturbation threshold :math:`\epsilon` in input space. Default: :math:`\frac{8}{255}`.

        norm (int): :math:`L_p` norm passed to :func:`torch.norm`. Default: ``float(inf)``.
        universal (bool): All inputs in the batch share the same noise. Default: ``False``.

        grad_method (str): gradient estimation method. Default: ``white``.
        query_num (int): number of samples in black box gradient estimation. Default: ``100``.
        sigma (float): gaussian noise std in black box gradient estimation. Default: ``0.001``.
    """

    name: str = 'pgd'

    def __init__(self, alpha: float = 3.0 / 255, epsilon: float = 8.0 / 255,
                 norm: Union[int, float] = float('inf'), universal: bool = False,
                 grad_method: str = 'white', query_num: int = 100, sigma: float = 1e-3,
                 hess_b: int = 100, hess_p: int = 1, hess_lambda: float = 1, **kwargs):
        super().__init__(**kwargs)
        self.param_list['pgd'] = ['alpha', 'epsilon', 'norm', 'universal']

        self.alpha = alpha
        self.epsilon = epsilon

        self.norm = norm
        self.universal = universal

        self.grad_method: str = grad_method
        if grad_method != 'white':
            self.param_list['blackbox'] = ['grad_method', 'query_num', 'sigma']
            self.query_num: int = query_num
            self.sigma: float = sigma
            if grad_method == 'hess':
                self.param_list['hessian'] = ['hess_b', 'hess_p', 'hess_lambda']
                self.hess_b: int = hess_b
                self.hess_p: int = hess_p
                self.hess_lambda: float = hess_lambda

    def optimize(self, _input: torch.Tensor, noise: torch.Tensor = None,
                 alpha: float = None, epsilon: float = None,
                 iteration: int = None, loss_fn: Callable[[torch.Tensor], torch.Tensor] = None,
                 output: Union[int, list[str]] = None, add_noise_fn=None, **kwargs) -> tuple[torch.Tensor, int]:
        # ------------------------------ Parameter Initialization ---------------------------------- #

        alpha = alpha if alpha is not None else self.alpha
        epsilon = epsilon if epsilon is not None else self.epsilon
        iteration = iteration if iteration is not None else self.iteration
        loss_fn = loss_fn if loss_fn is not None else self.loss_fn
        add_noise_fn = add_noise_fn if add_noise_fn is not None else add_noise
        noise = noise if noise is not None else torch.zeros_like(_input[0] if self.universal else _input)
        output = self.get_output(output)

        # ----------------------------------------------------------------------------------------- #

        if 'start' in output:
            self.output_info(_input=_input, noise=noise, mode='start', loss_fn=loss_fn, **kwargs)
        if iteration == 0 or alpha == 0.0 or epsilon == 0.0:
            return _input, None

        X = add_noise_fn(_input=_input, noise=noise, batch=self.universal)

        # ----------------------------------------------------------------------------------------- #

        for _iter in range(iteration):
            if self.early_stop_check(float(loss_fn(X))):
                if 'end' in output:
                    self.output_info(_input=_input, noise=noise, mode='end', loss_fn=loss_fn, **kwargs)
                return X.detach(), _iter + 1
            if self.grad_method == 'hess' and _iter % self.hess_p == 0:
                self.hess = self.calc_hess(loss_fn, X, sigma=self.sigma,
                                           hess_b=self.hess_b, hess_lambda=self.hess_lambda)
                self.hess /= self.hess.norm(p=2)
            grad = self.calc_grad(loss_fn, X)
            if self.grad_method != 'white' and 'middle' in output:
                real_grad = self.whitebox_grad(loss_fn, X)
                prints('cos<real, est> = ', cos_sim(grad.sign(), real_grad.sign()),
                       indent=self.indent + 2)
            if self.universal:
                grad = grad.mean(dim=0)
            noise.data = (noise - alpha * torch.sign(grad)).data
            noise.data = self.projector(noise, epsilon, norm=self.norm).data
            X = add_noise_fn(_input=_input, noise=noise, batch=self.universal)
            if self.universal:
                noise.data = (X - _input).mode(dim=0)[0].data
            else:
                noise.data = (X - _input).data

            if 'middle' in output:
                self.output_info(_input=_input, noise=noise, mode='middle',
                                 _iter=_iter, iteration=iteration, loss_fn=loss_fn, **kwargs)
        if 'end' in output:
            self.output_info(_input=_input, noise=noise, mode='end', loss_fn=loss_fn, **kwargs)
        return X.detach(), None

    def output_info(self, _input: torch.Tensor, noise: torch.Tensor, loss_fn=None, **kwargs):
        super().output_info(**kwargs)
        with torch.no_grad():
            loss = float(loss_fn(_input + noise))
            norm = noise.norm(p=self.norm)
            prints(f'L-{self.norm} norm: {norm}    loss: {loss:.5f}', indent=self.indent)

    @staticmethod
    def projector(noise: torch.Tensor, epsilon: float, norm: Union[float, int, str] = float('inf')) -> torch.Tensor:
        length = epsilon / noise.norm(p=norm)
        if length < 1:
            if norm == float('inf'):
                noise = noise.clamp(min=-epsilon, max=epsilon)
            else:
                noise = length * noise
        return noise

    # -------------------------- Calculate Gradient ------------------------ #
    def calc_grad(self, f, X: torch.Tensor) -> torch.Tensor:
        if self.grad_method != 'white':
            return self.blackbox_grad(f, X, query_num=self.query_num, sigma=self.sigma)
        else:
            return self.whitebox_grad(f, X)

    @staticmethod
    def whitebox_grad(f, X: torch.Tensor) -> torch.Tensor:
        X.requires_grad_()
        loss = f(X)
        grad = torch.autograd.grad(loss, X)[0]
        X.requires_grad = False
        return grad

    def blackbox_grad(self, f: Callable[[torch.Tensor], torch.Tensor], X: torch.Tensor) -> torch.Tensor:
        seq = self.gen_seq(X)
        grad = self.calc_seq(f, seq)
        return grad

    # X: (1, C, H, W)
    # return: (query_num+1, C, H, W)
    def gen_seq(self, X: torch.Tensor, query_num: int = None) -> torch.Tensor:
        query_num = query_num if query_num is not None else self.query_num
        sigma = self.sigma
        shape = list(X.shape)
        shape[0] = query_num
        if self.grad_method == 'nes':
            shape[0] = shape[0] // 2
        noise = sigma * torch.normal(mean=0.0, std=1.0, size=shape, device=X.device)

        zeros = torch.zeros_like(X)
        seq = [zeros]
        if self.grad_method == 'nes':
            seq.extend([noise, -noise])
            if query_num % 2 == 1:
                seq.append(zeros)
        elif self.grad_method == 'sgd':
            seq.append(noise)
        elif self.grad_method == 'hess':
            noise = (self.hess @ noise.view(-1, 1)).view(X.shape)
            seq.append(noise)
        elif self.grad_method == 'zoo':
            raise NotImplementedError(self.grad_method)
        else:
            print('Current method: ', self.grad_method)
            raise ValueError("Argument 'method' should be 'nes', 'sgd' or 'hess'!")
        seq = torch.cat(seq).add(X)
        return seq

    def calc_seq(self, f: Callable[[torch.Tensor], torch.Tensor], seq: torch.Tensor) -> torch.Tensor:
        X = seq[0].unsqueeze(0)
        seq = seq[1:]
        noise = seq.sub(X)
        with torch.no_grad():
            g = f(seq, reduction='none')[:, None, None, None].mul(noise).sum(dim=0)
            if self.grad_method in ['sgd', 'hess']:
                g -= f(X) * noise.sum(dim=0)
            g /= len(seq) * self.sigma * self.sigma
        return g

    @staticmethod
    def calc_hess(f: Callable[[torch.Tensor], torch.Tensor], X: torch.Tensor,
                  sigma: float, hess_b: int, hess_lambda: float = 1) -> torch.Tensor:
        length = X.numel()
        hess: torch.Tensor = torch.zeros(length, length, device=X.device)
        with torch.no_grad():
            for i in range(hess_b):
                noise = torch.normal(mean=0.0, std=1.0, size=X.shape, device=X.device)
                X1 = X + sigma * noise
                X2 = X - sigma * noise
                hess += abs(f(X1) + f(X2) - 2 * f(X)) * \
                    (noise.view(-1, 1) @ noise.view(1, -1))
            hess /= (2 * hess_b * sigma * sigma)
            hess += hess_lambda * torch.eye(length, device=X.device)
            result = hess.cholesky_inverse()
        return result



#!/usr/bin/env python3

from trojanzoo.utils.output import prints, ansi
from trojanzoo.utils.param import Module, Param

import os
import json
import yaml

from typing import TYPE_CHECKING
from typing import Any, Union    # TODO: python 3.10
ConfigFileType = Module[str, Union[Any, Param[str, Any]]]    # config_dict['package']['dataset'] dataset.yml
ConfigType = Module[str, ConfigFileType]                     # config_dict['package']
if TYPE_CHECKING:
    pass    # TODO: python 3.10


config_path: dict[str, str] = {
    'package': os.path.dirname(__file__),
    'user': None,
    'project': os.path.normpath('./configs/trojanzoo'),
}


class Config:
    name = 'config'
    cmd: str = None
    cmd_config: ConfigType = None

    @classmethod
    def update_cmd(cls, cmd_path: str = None) -> ConfigType:
        if cmd_path is not None and cmd_path != cls.cmd:
            if not os.path.exists(cmd_path):
                raise FileNotFoundError(cmd_path)
            cls.cmd = os.path.normpath(cmd_path)
            config = cls.load_config(cls.cmd)
            cls.cmd_config = config if len(config) else cls.cmd_config

    def __init__(self, _base=None, **kwargs: str):
        self.config_path = kwargs
        # self._base = _base
        self.config_dict: dict[str, ConfigType] = {}
        for key in self.config_path.keys():
            value = self.load_config(self.config_path[key])
            if len(value):
                self.config_dict[key] = value
        if _base is not None:
            self.config_dict = self.combine_base(self.config_dict, _base)
        self.full_config = self.combine()
        self.cmd_updated: bool = False

    def get_full_config(self):
        if not self.cmd_updated and self.cmd_config is not None:
            self.full_config.update(self.cmd_config)
            self.cmd_updated = True
        return self.full_config

    def get_config(self, dataset_name: str, config: ConfigType = None, **kwargs) -> Param[str, Module[str, Any]]:
        config = config if config is not None else Param(self.get_full_config(), default=Module())
        # remove dataset_name Param
        for file_name, file_value in config.items():
            if not isinstance(file_value, Module) and not isinstance(file_value, dict):
                # TODO: remove the latter condition?
                continue
            if isinstance(file_value, Param):
                config[file_name] = file_value[dataset_name]
                continue
            for param_name, param_value in file_value.items():
                if isinstance(param_value, Param):
                    config[file_name][param_name] = param_value[dataset_name]
                # else:
                #     raise TypeError(f'{type(param_value)=}    {param_value=}')
        config.update(kwargs)
        return config

    def combine(self, keys: list[str] = ['package', 'user', 'project']) -> ConfigType:
        config = Module()
        for key in keys:
            if key in self.config_dict.keys():
                config.update(self.config_dict[key])
        return config

    @staticmethod
    def combine_base(config_dict: dict[str, ConfigType], _base):
        _base: Config = _base
        for key, value in _base.items():
            value = value.copy()
            if key in config_dict.keys():
                value.update(config_dict[key])
            config_dict[key] = value
        return config_dict

    @staticmethod
    def load_config(path: str) -> ConfigType:
        if path is None:
            return {}
        elif not isinstance(path, str):     # TODO: unnecessary
            raise TypeError(path)
        elif not os.path.exists(path):
            return Module()
        elif os.path.isdir(path):
            _dict: ConfigType = Module()
            for root, dirs, files in os.walk(path):
                for _file in files:
                    name, ext = os.path.splitext(_file)
                    file_path = os.path.normpath(os.path.join(root, _file))
                    assert name not in _dict.keys(), f'filename conflicts: {file_path}'
                    if ext in ['.yml', '.yaml', 'json']:
                        _dict.update(Config.load_config(file_path))
            return _dict
        elif os.path.isfile(path):
            name, ext = os.path.splitext(os.path.split(path)[1])
            if ext in ['.yml', 'yaml', 'json']:
                with open(path, 'r', encoding='utf-8') as f:
                    _dict: dict[str, Union[Any, dict[str, Any]]] = {}
                    if ext == 'json':
                        _dict = json.load(f.read())
                    else:
                        _dict = yaml.load(f.read(), Loader=yaml.FullLoader)
                    return Module(**{name: Config.organize_config_file(_dict)})
            else:
                return Module()
        else:
            raise Exception(f'unknown: {path}')

    @staticmethod
    def organize_config_file(_dict: dict[str, Union[Any, dict[str, Any]]]) -> ConfigFileType:
        module = Module()
        for key, value in _dict.items():
            if isinstance(value, dict):
                value = Param(value)
            module[key] = value  # TODO: Shall we Param(value) ?
        return module

    def __getitem__(self, k: str):
        return self.config_dict[k]

    def items(self):
        return self.config_dict.items()

    def keys(self):
        return self.config_dict.keys()

    def summary(self, keys: Union[list[str], str] = None, config: ConfigType = None, indent: int = 0):
        if keys is None:
            prints('{yellow}{0:<20s}{reset} '.format(self.name, **ansi), indent=indent)
            self.summary(keys='final', config=self.get_full_config(), indent=indent + 10)
        elif isinstance(keys, list):
            prints('{yellow}{0:<20s}{reset} '.format(self.name, **ansi), indent=indent)
            for key in keys:
                if key in self.config_dict.items():
                    config = self.config_dict[key]
                elif key == 'cmd':
                    config = self.cmd_config
                elif key == 'final':
                    config = self.get_full_config()
                else:
                    raise KeyError(key)
                self.summary(keys=key, config=config, indent=indent + 10)
        else:
            assert isinstance(keys, str) and config is not None
            prints('{green}{0:<20s}{reset}'.format(keys, **ansi), indent=indent)
            for key, value in config.items():
                prints('{blue_light}{0:<20s}{reset}'.format(key, **ansi), indent=indent + 10)
                prints(value, indent=indent + 10)
                prints('-' * 20, indent=indent + 10)
            prints('-' * 30, indent=indent)

    def __str__(self) -> str:
        return str(self.config_dict)

    def __repr__(self):
        return repr(self.config_dict)


config = Config(**config_path)



#!/usr/bin/env python3

from trojanzoo.attacks import Attack
from trojanzoo.utils.output import prints

import torch
import numpy as np
import math
import random
import os
import argparse
from typing import Callable


class PoisonBasic(Attack):

    name: str = 'poison_basic'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--poison_percent', dest='poison_percent', type=float,
                           help='malicious training data injection probability for each batch, defaults to 0.1')
        group.add_argument('--target_idx', dest='target_idx', type=int,
                           help='Target label order in original classification, defaults to 1 '
                           '(0 for untargeted attack, 1 for most possible class, -1 for most unpossible class)')

    def __init__(self, poison_percent: float = 0.5, target_idx=1, **kwargs):
        super().__init__(**kwargs)
        self.param_list['poison'] = ['poison_percent', 'poison_num', 'target_idx']
        self.poison_percent: float = poison_percent
        self.target_idx: int = target_idx

        self.temp_input: torch.Tensor = None
        self.temp_label: torch.Tensor = None

        self.poison_num = self.dataset.batch_size * self.poison_percent

    def attack(self, epoch: int, **kwargs):
        # model._validate()
        total = 0
        target_conf_list = []
        target_acc_list = []
        clean_acc_list = []
        for data in self.dataset.loader['test']:
            if total >= 100:
                break
            self.model.load()
            _input, _label = self.model.remove_misclassify(data)
            if len(_label) == 0:
                continue
            _label = self.model.generate_target(_input, idx=self.target_idx)
            self._train(_input=_input, _label=_label, epoch=epoch, **kwargs)
            target_conf, target_acc, clean_acc = self.validate_fn()
            target_conf_list.append(target_conf)
            target_acc_list.append(target_acc)
            clean_acc_list.append(clean_acc)
            total += 1
            print(f'[{total+1}/100]\n'
                  f'target confidence: {np.mean(target_conf_list)}({np.std(target_conf_list)})\n'
                  f'target accuracy: {np.mean(target_acc_list)}({np.std(target_acc_list)})\n'
                  f'clean accuracy: {np.mean(clean_acc_list)}({np.std(clean_acc_list)})\n\n\n')

    def _train(self, _input: torch.Tensor, _label: torch.Tensor, epoch: int, save=False, indent=0, **kwargs):
        self.temp_input = _input
        self.temp_label = _label

        self.model._train(epoch=epoch, save=save,
                          get_data_fn=self.get_data, save_fn=self.save,
                          validate_fn=self.validate_fn, indent=indent + 4, **kwargs)

    def get_data(self, data: tuple[torch.Tensor, torch.Tensor], keep_org: bool = True, poison_label=True, **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        _input, _label = self.model.get_data(data)
        decimal, integer = math.modf(self.poison_num)
        integer = int(integer)
        if random.uniform(0, 1) < decimal:
            integer += 1
        if not keep_org or integer:
            org_input, org_label = _input, _label
            _input = torch.cat([self.temp_input] * integer)
            if poison_label:
                _label = torch.cat([self.temp_label] * integer)
            if keep_org:
                _input = torch.cat((_input, org_input))
                _label = torch.cat((_label, org_label))
        return _input, _label

    # def loss_fn(self, x: torch.Tensor, y: torch.Tensor, **kwargs):
    #     clean = self.model.loss(x, y, **kwargs)
    #     training: bool = self.model._model.training
    #     self.model.eval()
    #     poison = self.model.loss(self.temp_input, self.temp_label)
    #     self.model.train(mode=training)
    #     print(f'clean: {clean:7.5f}    poison: {poison:7.5f}   eval: {poison:7.5f}')
    #     return (1 - self.poison_percent) * self.model.loss(x, y, **kwargs) + self.poison_percent * self.model.loss(self.temp_input, self.temp_label)

    def save(self, **kwargs):
        filename = self.get_filename(**kwargs)
        file_path = os.path.join(self.folder_path, filename)
        self.model.save(file_path + '.pth')
        print('attack results saved at: ', file_path)

    def get_filename(self, **kwargs):
        return self.model.name

    def validate_target(self, indent: int = 0, verbose=True) -> tuple[float, float]:
        self.model.eval()
        _output = self.model(self.temp_input)
        target_acc, _ = self.model.accuracy(_output, self.temp_label, topk=(1, 5))
        target_conf = float(self.model.get_target_prob(self.temp_input, self.temp_label).mean())
        target_loss = self.model.loss(self.temp_input, self.temp_label)
        if verbose:
            prints(f'Validate Target:       Loss: {target_loss:10.4f}     Confidence: {target_conf:10.4f}    Accuracy: {target_acc:7.3f}',
                   indent=indent)
        # todo: Return value
        return target_conf, target_acc

    def validate_fn(self, get_data_fn: Callable[..., tuple[torch.Tensor, torch.Tensor]] = None,
                    main_tag: str = 'valid', indent: int = 0, verbose=True, **kwargs) -> tuple[float, float]:
        _, target_acc = self.validate_target(indent=indent, verbose=verbose)
        _, clean_acc = self.model._validate(print_prefix='Validate Clean', main_tag='valid clean',
                                            get_data_fn=None, indent=indent, **kwargs)
        return clean_acc, target_acc



#!/usr/bin/env python3

from ..backdoor_defense import BackdoorDefense
from trojanvision.environ import env
from trojanzoo.utils.data import TensorListDataset

import torch
from sklearn.decomposition import FastICA, PCA
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score
from sklearn.metrics import f1_score
from sklearn import metrics
import argparse
from tqdm import tqdm


class ActivationClustering(BackdoorDefense):

    """
    Activation Clustering Defense is described in the paper 'Detecting Backdoor Attacks on Deep Neural Networks by Activation Clustering'_ by Bryant Chen. The main idea is the reason why backdoor and target samples receive the same classification is different, the network identifies features in the input taht it has learned corresponding to the target class, in the case of backdoor samples, it identifies features associated with the source class and the backdoor trigger. This difference in mechanism is evident in the network activations.

    The authors have posted `original source code`_. This defence is intergrated into the adversarial-robustness-toolbox provided by IBM.

    Args:
        mix_image_num (int): the number of sampled image, including clean image and poison image. Default: 50.
        clean_image_ratio (float): the ratio of clean image, compared to mix_image_num. Default: 0.5.
        retrain_epoch (int): the epoch of retraining the model, only with the detected clean image. Default: 10.
        nb_clusters (int): the amount of clusters. Default: 2. 
        clustering_method (str): the method for clustering the data, only support KMeans. Default: 'KMeans'.
        nb_dims (int): the dimension set in the process of reduceing the dimensionality of data. Default: 10.
        reduce_method (str): the method for reduing the dimensionality of data, only supporting ICA and PCA. Default: FastICA.
        cluster_analysis (str): the method chosen to analyze whether cluster is the poison cluster, including size, relative-size,exclusionary-reclassification, silhouette-scores. Default: size.



    .. _Activation Clustering:
        https://arxiv.org/abs/1811.03728

    .. _original source code:
        https://github.com/IBM/adversarial-robustness-toolbox


    Raises:
        ValueError: dimensionality reduction method not supported.
        ValueError: clustering method not supported.
        ValueError: *** Analyzer does not support more than two clusters.
        ValueError: Unsupported cluster analysis technique.
        ValueError: SC score too low, the sample should be contained in one cluster.

    Returns:
        model(ImageModel): the clean model trained only with clean samples.
    """

    name: str = 'activation_clustering'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--mix_image_num', dest='mix_image_num', type=int,
                           help='the number of sampled image')
        group.add_argument('--clean_image_ratio', dest='clean_image_ratio', type=float,
                           help='the ratio of clean image')
        group.add_argument('--retrain_epoch', dest='retrain_epoch', type=int,
                           help='the epoch of retraining the model')
        group.add_argument('--nb_clusters', dest='nb_clusters', type=int,
                           help='')
        group.add_argument('--clustering_method', dest='clustering_method', type=str,
                           help='the amount of clusters')
        group.add_argument('--nb_dims', dest='nb_dims', type=int,
                           help='the dimension set in the process of reduceing the dimensionality of data')
        group.add_argument('--reduce_method', dest='reduce_method', type=str,
                           help=' the method for reduing the dimensionality of data')
        group.add_argument('--cluster_analysis', dest='cluster_analysis', type=str,
                           help='the method chosen to analyze whether cluster is the poison cluster, including size, distance, relative-size, silhouette-scores')

    def __init__(self, mix_image_num: int = 100, clean_image_ratio: float = 0.95, retrain_epoch: int = 10, nb_clusters: int = 2, clustering_method: str = "KMeans", nb_dims: int = 10, reduce_method: str = "FastICA", cluster_analysis: str = "exclusionary-reclassification", **kwargs):
        super().__init__(**kwargs)

        self.mix_image_num = mix_image_num
        self.clean_image_ratio = clean_image_ratio
        self.clean_image_num = int(mix_image_num * clean_image_ratio)
        self.poison_image_num = self.mix_image_num - self.clean_image_num

        self.nb_clusters = nb_clusters
        self.clustering_method = clustering_method
        self.nb_dims = nb_dims
        self.reduce_method = reduce_method
        self.cluster_analysis = cluster_analysis

        self.retrain_epoch = retrain_epoch

        self.clean_dataset, _ = self.dataset.split_set(
            self.dataset.get_full_dataset(mode='train'), self.clean_image_num)
        # clean_dataset, _ = self.dataset.split_set(self.dataset.get_full_dataset(mode='train'), self.clean_image_num)
        # clean_dataloader = self.dataset.get_dataloader(mode='train', dataset=clean_dataset, batch_size=self.clean_image_num, num_workers=0)
        # clean_imgs, _ = self.model.get_data(next(iter(clean_dataloader)))
        # self.clean_dataset = TensorDataset(clean_imgs, _)

        poison_dataset, _ = self.dataset.split_set(self.dataset.get_full_dataset(mode='train'), self.poison_image_num)
        poison_dataloader = self.dataset.get_dataloader(
            mode='train', dataset=poison_dataset, batch_size=self.poison_image_num, num_workers=0)
        poison_imgs, _ = self.model.get_data(next(iter(poison_dataloader)))
        poison_imgs = self.attack.add_mark(poison_imgs).cpu()
        poison_label = [self.attack.target_class] * self.poison_image_num
        self.poison_dataset = TensorListDataset(poison_imgs, poison_label)
        self.mix_dataset = torch.utils.data.ConcatDataset([self.clean_dataset, self.poison_dataset])
        self.mix_dataloader = self.dataset.get_dataloader(
            mode='train', dataset=self.mix_dataset, batch_size=self.dataset.batch_size, num_workers=0)

    def detect(self, optimizer, lr_scheduler, **kwargs):
        """
        Record the detected poison samples
        Remove them and retrain the model from scratch to get a clean model.
        """
        super().detect(**kwargs)
        all_reduced_activations, all_label, all_clusters, all_class = self.preprocess(self.mix_dataloader)
        poison_cluster_index = self.analyze_clusters(
            all_clusters, all_reduced_activations, all_label, all_class, self.cluster_analysis)
        print(all_class)
        poison_input_index = []
        for i in range(len(all_clusters)):
            if all_clusters[i] == poison_cluster_index:
                poison_input_index.append(i)

        y_pred = torch.zeros(self.mix_image_num)
        for i in range(len(poison_input_index)):
            y_pred[poison_input_index[i]] = 1

        y_true = torch.zeros(self.mix_image_num)
        for i in range(self.clean_image_num, self.mix_image_num):
            y_true[i] = 1

        # print("y_pred: ", y_pred)
        # print("y_true: ", y_true)

        final_f1_score = f1_score(y_true, y_pred, average='weighted')
        print("f1_score:", final_f1_score)
        print("precision_score:", metrics.precision_score(y_true, y_pred, average='weighted'))
        print("recall_score:", metrics.recall_score(y_true, y_pred, average='weighted'))
        print("accuracy_score:", metrics.accuracy_score(y_true, y_pred))

    def preprocess(self, loader):
        """
        Get the feature map of samples, convert to 1D tensor, reduce the dimensionality and cluster them.

        Args:
            loader (torch.utils.data.dataloader): the mix dataloader of clean image and poison image

        Returns:
            all_label (torch.Tensor): the prediction label of the model on all_input 
            all_reduced_activation (torch.FloatTensor): the reduced activation of all input getting from the model
            all_clusters (torch.Tensor): the clustering result
        """

        all_feature_map = []
        all_label = []
        all_class = []
        if env['tqdm']:
            loader = tqdm(loader)
        for i, data in enumerate(loader):
            _input, _label = self.model.get_data(data)
            feature_map = self.model._model.get_fm(_input)
            pred_label = self.model.get_class(_input)
            all_feature_map.append(feature_map)
            all_label.append(pred_label)
            all_class.append(_label)
        all_feature_map = torch.cat(all_feature_map)
        all_label = torch.cat(all_label)
        all_class = torch.cat(all_class)
        all_reduced_activations = self.reduce_dimensionality(all_feature_map, self.nb_dims, self.reduce_method)
        all_clusters = self.cluster_activations(all_reduced_activations, self.nb_clusters, self.clustering_method)
        all_clusters = torch.Tensor(all_clusters)
        return all_reduced_activations, all_label, all_clusters, all_class

    def reduce_dimensionality(self, activations, nb_dims: int = 10, reduce_method: str = "FastICA"):
        """
        Reduce dimensionality of activations.

        Args:
            activations (torch.FloatTensor): [description]
            nb_dims (int, optional): the setted dimensionality after reducing dimensionality. Defaults to 10.
            reduce_method (str, optional): the chosen reducing dimensionality method. Defaults to "FastICA".

        Raises:
            ValueError: dimensionality reduction method not supported.

        Returns:
            reduced_activations(torch.FloatTensor): the result after reduing dimensionality.
        """
        if reduce_method == "FastICA":
            projector = FastICA(n_components=nb_dims)
        elif reduce_method == "PCA":
            projector = PCA(n_components=nb_dims)
        else:
            raise ValueError(reduce_method + " dimensionality reduction method not supported.")
        activations = activations.view(activations.shape[0], -1)
        reduced_activations = projector.fit_transform(activations.detach().cpu())
        return reduced_activations

    def cluster_activations(self, reduced_activations, nb_clusters: int = 2, clustering_method: str = "KMeans"):
        """
        Cluster the activations after reducing dimensionality.

        Args:
            reduced_activations (torch.FloatTensor): the result after reduing dimensionality.
            nb_clusters (int, optional): the amount of clusters. Defaults to 2.
            clustering_method (str, optional): the chosen clustering method. Defaults to "KMeans".

        Raises:
            ValueError: clustering method not supported.

        Returns:
            clusters(torch.Tensor): the result of clustering
        """
        if clustering_method == "KMeans":
            clusterer = KMeans(n_clusters=nb_clusters)
            clusters = clusterer.fit_predict(reduced_activations)
            return clusters
        else:
            raise ValueError(clustering_method + " clustering method not supported.")

    def analyze_by_size(self, cluster_pred):
        """
        Analyze the result of clustering to judge which cluster is poison, according the size of clusters, usually the poisoned sample is less compared with normal data, so the smaller cluster will be regareded as poisoned samples.

        Args:
            cluster_pred (torch.Tensor): the result of clustering

        Raises:
            ValueError: Size Analyzer does not support more than two clusters.

        Returns:
            poison_cluster_index : the poisoned cluster number.
        """
        if (len(torch.unique(cluster_pred)) > 2):
            raise ValueError("Size Analyzer does not support more than two clusters.")
        num_1 = 0
        num_1 = torch.sum(cluster_pred)
        if num_1 > len(cluster_pred) / 2:
            poison_cluster_index = 0
            return poison_cluster_index
        else:
            poison_cluster_index = 1
            return poison_cluster_index

    def analyze_by_relative_size(self, label, cluster_pred, size_threshold: float = 0.35):
        """
        Analyze the result of clustering to judge which cluster is poison, according the ith label cluster result, if the portion of the smaller cluster is lower than size_threshold, this cluster is regarded as poison.

        Args:
            label (torch.Tensor): the original label of data in mix_dataloader.
            cluster_pred (torch.Tensor): the result of clustering
            size_threshold (float, optional): experience value. Defaults to 0.35.

        Raises:
            ValueError: Relative_Size Analyzer does not support more than two clusters.

        Returns:
            poison_cluster_index : the poisoned cluster number.
        """
        if (len(torch.unique(cluster_pred)) > 2):
            raise ValueError("Relative_Size Analyzer does not support more than two clusters.")
        else:
            all_classes = list(range(self.dataset.num_classes))
            for i in range(len(all_classes)):
                cur_label = all_classes[i]
                cur_label_num = torch.bincount(label)[cur_label]
                zip_label = list(zip(label, cluster_pred))
                num_0_cur_label = 0
                for j in range(len(zip_label)):
                    if list(zip_label[j])[0] == cur_label and list(zip_label[j])[1] == 0:
                        num_0_cur_label += 1
                if float(num_0_cur_label / cur_label_num) < size_threshold:
                    poison_cluster_index = 0
                    return poison_cluster_index
                elif float((cur_label_num - num_0_cur_label) / cur_label_num) < size_threshold:
                    poison_cluster_index = 1
                    return poison_cluster_index
                else:
                    continue

    def analyze_by_exclusionary_reclassification(self, cluster_pred, label, all_class):
        """
        Analyze the result of clustering to judge which cluster is poison, according the class of most data and the number of data of which the predicted and true class is identical. 

        Args:
            cluster_pred (torch.Tensor): the result of clustering
            label (torch.Tensor): the original label of data in mix_dataloader.
            all_class (torch.Tensor): the true class of sampled data.

        Returns:
            poison_cluster_index : the poisoned cluster number.
        """
        zip_label = list(zip(cluster_pred, label, all_class))
        # print('true label:',label)
        # print('predicted class:',all_class)
        max_label = torch.argmax(torch.bincount(label))
        print("max_label:{}".format(max_label))
        class_0_correct_num = 0
        class_1_correct_num = 0
        class_0_max_num = 0
        class_1_max_num = 0

        for i in range(len(zip_label)):
            if list(zip_label[i])[0] == 0:
                if list(zip_label[i])[1] == max_label:

                    class_0_max_num += 1
                if list(zip_label[i])[1] == list(zip_label[i])[2]:

                    class_0_correct_num += 1
            elif list(zip_label[i])[0] == 1:
                if list(zip_label[i])[1] == max_label:

                    class_1_max_num += 1
                if list(zip_label[i])[1] == list(zip_label[i])[2]:

                    class_1_correct_num += 1
        # try:
        if float(class_0_correct_num / (class_0_max_num + 1)) > float(class_1_correct_num / (class_1_max_num + 1)):
            poison_cluster_index = 1
            return poison_cluster_index
        else:
            poison_cluster_index = 0
            return poison_cluster_index
        # except ZeroDivisionError:
        #     print("Error: you can't divide by 0!")
        # else:
        #     poison_cluster_index = self.analyze_by_size(cluster_pred)
        #     return poison_cluster_index

    def analyze_by_silhouette_score(self, reduced_activation, cluster_pred, label, score_threshold: float = 0.1):
        """
        Analyze the result of clustering to judge which cluster is poison, according the silhouette_score, which specifies whether the number of clusters fits well with the data, the higher, the better. Test the situation under nb_clusters set as 2 and different class of data, if the silhouette_score is high, the smaller cluster will be the poison cluster.

        Args:
            reduced_activation (torch.FloatTensor): the reduced activation of sample in mix_dataloader.
            cluster_pred (torch.Tensor): the result of clustering.
            score_threshold (float, optional): experience value. Defaults to 0.1.

        Raises:
            ValueError: Silhouette_score Analyzer does not support more than two clusters.
            ValueError: SC score too low, the sample should be contained in one cluster.
            ValueError: clustering method not supported.

        Returns:
            poison_cluster_index : the poisoned cluster number.
        """
        if (len(torch.unique(cluster_pred)) > 2):
            raise ValueError("Silhouette_score Analyzer does not support more than two clusters.")
        else:
            if self.clustering_method == "KMeans":
                all_classes = list(range(self.dataset.num_classes))
                zip_label = list(zip(label, cluster_pred, reduced_activation))
                for i in range(len(all_classes)):
                    cur_label = all_classes[i]
                    cur_label_activation = []
                    cur_label_cluster_pred = []
                    for j in range(len(zip_label)):
                        if list(zip_label[j])[0] == cur_label:
                            cur_label_cluster_pred.append(list(zip_label[j])[1])
                            cur_label_activation.append(list(zip_label[j])[2])

                    if len(cur_label_activation) != 0:

                        cur_label_activation = torch.tensor(cur_label_activation)
                        cur_label_cluster_pred = torch.tensor(cur_label_cluster_pred)
                        kmeans_model = KMeans(n_clusters=self.nb_clusters).fit(cur_label_activation)
                        sc_score = silhouette_score(cur_label_activation, kmeans_model.labels_, metric='euclidean')
                        if sc_score > score_threshold:
                            poison_cluster_index = self.analyze_by_size(cur_label_cluster_pred)
                            return poison_cluster_index
                        else:
                            continue
                    else:
                        continue

            else:
                raise ValueError(self.clustering_method + " clustering method not supported.")

    def analyze_clusters(self, cluster_pred, reduced_activation, label, all_class, cluster_analysis: str = 'size', **kwargs):
        """
        Chooose the method of analyzing the clusters.

        Args:
            mix_feature_map (torch.FloatTensor): the feature map of sample in mix_dataloader.
            cluster_pred (torch.Tensor): the result of clustering.
            label (torch.Tensor): the original label of data in mix_dataloader.
            all_class (torch.Tensor): the true class of sampled data.
            cluster_analysis (str): the chosen cluster analyzing method.

        Returns:
            poison_cluster_index: the poisoned cluster number.

        """

        if cluster_analysis == "size":
            poison_cluster_index = self.analyze_by_size(cluster_pred)
        elif cluster_analysis == "relative-size":
            poison_cluster_index = self.analyze_by_relative_size(label, cluster_pred)
        elif cluster_analysis == "silhouette-scores":
            poison_cluster_index = self.analyze_by_silhouette_score(reduced_activation, cluster_pred, label)
        elif cluster_analysis == "exclusionary-reclassification":
            poison_cluster_index = self.analyze_by_exclusionary_reclassification(cluster_pred, label, all_class)
        else:
            raise ValueError("Unsupported cluster analysis technique " + cluster_analysis)
        return poison_cluster_index



#!/usr/bin/env python3

from .badnet import BadNet
from trojanvision.environ import env
from trojanzoo.utils import to_tensor
from trojanzoo.utils import AverageMeter
from trojanzoo.utils.output import prints

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import TensorDataset

import argparse
from collections import OrderedDict


class BypassEmbed(BadNet):
    name: str = 'bypass_embed'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--lambd', dest='lambd', type=int)
        group.add_argument('--discrim_lr', dest='discrim_lr', type=float)
        group.add_argument('--poison_num', dest='poison_num', type=int)

    def __init__(self, lambd: int = 10, discrim_lr: float = 0.001,
                 **kwargs):
        super().__init__(**kwargs)
        self.param_list['bypass_embed'] = ['lambd', 'discrim_lr']
        self.lambd = lambd
        self.discrim_lr = discrim_lr

    def attack(self, epoch: int, lr_scheduler: optim.lr_scheduler._LRScheduler = None,
               save: bool = False, **kwargs):
        print('Sample Data')
        poison_loader, discrim_loader = self.sample_data()  # with poisoned images
        print('Joint Training')
        super().attack(epoch=10, lr_scheduler=lr_scheduler, **kwargs)
        if isinstance(lr_scheduler, optim.lr_scheduler._LRScheduler):
            lr_scheduler.step(0)
        self.joint_train(epoch=epoch, poison_loader=poison_loader, discrim_loader=discrim_loader,
                         save=save, lr_scheduler=lr_scheduler, **kwargs)

    def sample_data(self):
        other_classes = list(range(self.dataset.num_classes))
        other_classes.pop(self.target_class)
        other_x, other_y = [], []
        poison_num = len(self.dataset.get_dataset('train')) * self.poison_percent / self.dataset.num_classes
        for _class in other_classes:
            loader = self.dataset.get_dataloader(mode='train', batch_size=int(poison_num), classes=[_class],
                                                 shuffle=True, num_workers=0, pin_memory=False)
            _input, _label = next(iter(loader))
            other_x.append(_input)
            other_y.append(_label)
        other_x = torch.cat(other_x)
        other_y = torch.cat(other_y)

        poison_x = self.mark.add_mark(other_x)
        poison_y = self.target_class * torch.ones_like(other_y)

        trainset = self.dataset.get_dataset(mode='train')
        clean_x, clean_y = zip(*trainset)
        clean_x = torch.stack(clean_x)
        clean_y = torch.tensor(clean_y)

        discrim_x = torch.cat((other_x, poison_x))
        discrim_y = torch.cat((torch.zeros_like(other_y),
                               torch.ones_like(poison_y)))
        discrim_set = TensorDataset(discrim_x, discrim_y)
        discrim_loader = self.dataset.get_dataloader(
            mode='train', dataset=discrim_set, batch_size=self.dataset.batch_size)

        all_x = torch.cat((clean_x, poison_x))
        all_y = torch.cat((clean_y, poison_y))
        all_discrim_y = torch.cat((torch.zeros_like(clean_y),
                                   torch.ones_like(poison_y)))
        # used for training
        poison_set = TensorDataset(all_x, all_y, all_discrim_y)
        poison_loader = self.dataset.get_dataloader(mode='train', dataset=poison_set,
                                                    batch_size=self.dataset.batch_size)
        return poison_loader, discrim_loader

    @staticmethod
    def bypass_get_data(data: tuple[torch.Tensor], **kwargs) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        return to_tensor(data[0]), to_tensor(data[1], dtype='long'), to_tensor(data[2], dtype='long')

    def joint_train(self, epoch: int = 0, optimizer: optim.Optimizer = None, lr_scheduler: optim.lr_scheduler._LRScheduler = None,
                    poison_loader=None, discrim_loader=None, save=False, **kwargs):
        in_dim = self.model._model.classifier[0].in_features
        D = nn.Sequential(OrderedDict([
            ('fc1', nn.Linear(in_dim, 256)),
            ('bn1', nn.BatchNorm1d(256)),
            ('relu1', nn.LeakyReLU()),
            ('fc2', nn.Linear(256, 128)),
            ('bn2', nn.BatchNorm1d(128)),
            ('relu2', nn.ReLU()),
            ('fc3', nn.Linear(128, 2))
        ]))
        if env['num_gpus']:
            D.cuda()
        optim_params = [param_group['params'] for param_group in optimizer.param_groups]
        optimizer.zero_grad()

        best_acc = 0.0
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')

        for _epoch in range(epoch):
            self.discrim_train(epoch=100, D=D, discrim_loader=discrim_loader)

            self.model.train()
            self.model.activate_params(optim_params)
            for data in poison_loader:
                optimizer.zero_grad()
                _input, _label_f, _label_d = self.bypass_get_data(data)
                out_f = self.model(_input)
                loss_f = self.model.criterion(out_f, _label_f)
                out_d = D(self.model.get_final_fm(_input))
                loss_d = self.model.criterion(out_d, _label_d)

                loss = loss_f - self.lambd * loss_d
                loss.backward()
                optimizer.step()
                if lr_scheduler:
                    lr_scheduler.step()
                optimizer.zero_grad()
            self.model.activate_params([])
            self.model.eval()
            _, cur_acc = self.validate_fn(get_data_fn=self.bypass_get_data)
            if cur_acc >= best_acc:
                prints('best result update!', indent=0)
                prints(f'Current Acc: {cur_acc:.3f}    Previous Best Acc: {best_acc:.3f}', indent=0)
                best_acc = cur_acc
                if save:
                    self.save()
            print('-' * 50)

    def discrim_train(self, epoch: int, D: nn.Sequential, discrim_loader: torch.utils.data.DataLoader):
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')
        d_optimizer = optim.Adam(D.parameters(), lr=self.discrim_lr)
        d_optimizer.zero_grad()
        for _epoch in range(epoch):
            losses.reset()
            top1.reset()
            self.model.activate_params([D.parameters()])
            D.train()
            for data in discrim_loader:
                # train D
                _input, _label = self.model.get_data(data)
                out_f = self.model.get_final_fm(_input).detach()
                out_d = D(out_f)
                loss_d = self.model.criterion(out_d, _label)

                acc1 = self.model.accuracy(out_d, _label, topk=(1, ))[0]
                batch_size = int(_label.size(0))
                losses.update(loss_d.item(), batch_size)
                top1.update(acc1, batch_size)

                loss_d.backward()
                d_optimizer.step()
                d_optimizer.zero_grad()
            print(f'Discriminator - epoch {_epoch:4d} / {epoch:4d} | loss {losses.avg:.4f} | acc {top1.avg:.4f}')
            self.model.activate_params([])
            D.eval()



#!/usr/bin/env python3

import io
import re
import sys


class ANSI:
    ansi_color = {
        'black': '\033[30m',
        'red': '\033[31m',
        'green': '\033[32m',
        'yellow': '\033[33m',
        'blue': '\033[34m',
        'purple': '\033[35m',
        'blue_light': '\033[36m',
        'white': '\033[37m',

        'reset': '\033[0m',
        'upline': '\033[1A',
        'clear_line': '\033[2K',
        'clear': '\033[2J', }
    ansi_nocolor = {
        'black': '',
        'red': '',
        'green': '',
        'yellow': '',
        'blue': '',
        'purple': '',
        'blue_light': '',
        'white': '',

        'reset': '',
        'upline': '\033[1A\033[',
        'clear_line': '\033[K',
        'clear': '\033[2J', }

    def __init__(self):
        self._dict = ANSI.ansi_color if ('--color' in sys.argv) else ANSI.ansi_nocolor

    def switch(self, color: bool):
        self._dict = ANSI.ansi_color if color else ANSI.ansi_nocolor

    def keys(self):
        return self._dict.keys()

    def items(self):
        return self._dict.items()

    def __getitem__(self, key):
        return self._dict[key]

    def __str__(self):
        return self._dict.__str__()

    def __repr__(self):
        return self._dict.__repr__()


ansi = ANSI()


def remove_ansi(s: str) -> str:
    ansi_escape = re.compile(r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]')
    return ansi_escape.sub('', s)


def get_ansi_len(s: str) -> int:
    return len(s) - len(remove_ansi(s))


def prints(*args: str, indent: int = 0, prefix: str = '', **kwargs):
    assert indent >= 0
    new_args = []
    for arg in args:
        new_args.append(indent_str(arg, indent=indent))
    new_args[0] = prefix + str(new_args[0])
    print(*new_args, **kwargs)


def output_iter(_iter: int, iteration: int = None) -> str:
    if iteration is None:
        return '{blue_light}[ {red}{0:s}{blue_light} ]{reset}'.format(str(_iter).rjust(3), **ansi)
    else:
        length = len(str(iteration))
        return '{blue_light}[ {red}{0:s}{blue_light} / {red}{1:d}{blue_light} ]{reset}'.format(
            str(_iter).rjust(length), iteration, **ansi)


def indent_str(s_: str, indent: int = 0) -> str:
    # modified from torch.nn.modules._addindent
    if indent == 0:
        return s_
    tail = ''
    s_ = str(s_)
    if s_[-1] == '\n':
        s_ = s_[:-1]
        tail = '\n'
    s = str(s_).split('\n')
    s = [(indent * ' ') + line for line in s]
    s = '\n'.join(s)
    s += tail
    return s


class Indent_Redirect:
    def __init__(self, buffer: bool = False, indent: int = 0):
        self.__console__: io.TextIOWrapper = sys.stdout
        self.indent: int = indent
        self.buffer: str = None
        if buffer:
            self.buffer = ''

    def write(self, text: str, indent: int = None):
        indent = indent if indent is not None else self.indent
        text = indent_str(text, indent=indent)
        if self.buffer is None:
            self.__console__.write(text)
        else:
            self.buffer += text

    def flush(self):
        if self.buffer:
            self.__console__.write(self.buffer)
            self.buffer = ''
        self.__console__.flush()

    def reset(self):
        if self.buffer:
            self.buffer = ''
        sys.stdout = self.__console__



#!/usr/bin/env python3

from ..backdoor_defense import BackdoorDefense
from trojanzoo.utils import normalize_mad
from trojanzoo.utils.output import output_iter

import torch
import torch.nn as nn
from torch.utils.data import TensorDataset, DataLoader

import argparse


class NeuronInspect(BackdoorDefense):

    name: str = 'neuron_inspect'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--lambd_sp', dest='lambd_sp', type=float,
                           help='control sparse feature')
        group.add_argument('--lambd_sm', dest='lambd_sm', type=float,
                           help='control smooth feature')
        group.add_argument('--lambd_pe', dest='lambd_pe', type=float,
                           help='control persistence feature')
        group.add_argument('--thre', dest='thre', type=float,
                           help='Threshold for calculating persistence feature')
        group.add_argument('--sample_ratio', dest='sample_ratio', type=float,
                           help='sample ratio from the full training data')

    def __init__(self, lambd_sp: float = 1e-5, lambd_sm: float = 1e-5, lambd_pe: float = 1,
                 thre: float = 0, sample_ratio: float = 0.1, **kwargs):
        super().__init__(**kwargs)

        self.param_list['neuron_inspect'] = ['lambd_sp', 'lambd_sm', 'lambd_pe', 'thre', 'sample_ratio']

        self.lambd_sp: float = lambd_sp
        self.lambd_sm: float = lambd_sm
        self.lambd_pe: float = lambd_pe

        self.thre: float = thre
        self.sample_ratio: float = sample_ratio

        kernel = torch.tensor([[0., 1., 0.],
                               [1., -4., 1.],
                               [0., 1., 0.]], device='cpu')
        self.conv2d = nn.Conv2d(1, 1, 3, bias=False)
        self.conv2d.weight = nn.Parameter(kernel.view_as(self.conv2d.weight))

    def detect(self, **kwargs):
        super().detect(**kwargs)
        exp_features = self.get_explation_feature()
        exp_features = torch.tensor(exp_features)
        print('exp features: ', exp_features)
        print('exp mad: ', normalize_mad(exp_features))

    def get_explation_feature(self) -> list[float]:
        dataset = self.dataset.get_dataset(mode='train')
        subset, _ = self.dataset.split_set(dataset, percent=self.sample_ratio)
        clean_loader = self.dataset.get_dataloader(mode='train', dataset=subset)

        _input, _label = zip(*subset)
        _input = torch.stack(_input)
        _label = torch.tensor(_label)
        poison_input = self.attack.add_mark(_input)
        newset = TensorDataset(poison_input, _label)
        backdoor_loader = self.dataset.get_dataloader(mode='train', dataset=newset)

        exp_features = []
        for label in range(self.model.num_classes):
            print('Class: ', output_iter(label, self.model.num_classes))
            backdoor_saliency_maps = self.saliency_map(label, backdoor_loader)   # (N, H, W)
            benign_saliency_maps = self.saliency_map(label, clean_loader)        # (N, H, W)
            exp_features.append(self.cal_explanation_feature(backdoor_saliency_maps, benign_saliency_maps))
        return exp_features

    def saliency_map(self, target: int, loader: DataLoader, method='saliency_map') -> torch.Tensor:
        saliency_maps = []
        for data in loader:
            _input, _label = self.model.get_data(data)
            saliency_map = self.model.get_heatmap(_input, [target] * len(_input),
                                                  method=method, cmap=None)
            saliency_maps.append(saliency_map)
        return torch.cat(saliency_maps)  # (N, H, W)

    def cal_explanation_feature(self, backdoor_saliency_maps: torch.Tensor,
                                benign_saliency_maps: torch.Tensor) -> float:
        sparse_feats = backdoor_saliency_maps.flatten(start_dim=1).norm(p=1, dim=1)  # (N)
        smooth_feats = self.conv2d(backdoor_saliency_maps.unsqueeze(1)).flatten(start_dim=1).norm(p=1, dim=1)  # (N)
        persist_feats = self.cal_persistence_feature(benign_saliency_maps)  # (1)

        exp_feats = self.lambd_sp * sparse_feats + self.lambd_sm * smooth_feats + self.lambd_pe * persist_feats
        return torch.median(exp_feats).item()

    def cal_persistence_feature(self, saliency_maps: torch.Tensor) -> torch.Tensor:
        self.thre = torch.median(saliency_maps).item()
        saliency_maps = torch.where(saliency_maps > self.thre, torch.tensor(1.0), torch.tensor(0.0))
        _base = saliency_maps[0]
        for i in range(1, len(saliency_maps)):
            _base = torch.logical_xor(_base, saliency_maps[i]).float()
        return _base.flatten(start_dim=1).norm(p=1)



#!/usr/bin/env python3

from .neural_cleanse import NeuralCleanse
from .tabor import TABOR
from .strip import STRIP
from .abs import ABS
from .fine_pruning import FinePruning
from .activation_clustering import ActivationClustering
from .deep_inspect import DeepInspect
from .spectral_signature import SpectralSignature
from .neuron_inspect import NeuronInspect
from .image_transform import ImageTransform
from .adv_train import AdvTrain
from .magnet import MagNet
from .neo import NEO



#!/usr/bin/env python3

from .imageset import ImageSet
from trojanvision.utils.data import MemoryDataset, ZipFolder
from trojanvision.environ import env
from trojanzoo.utils.output import ansi, prints, output_iter
from trojanzoo.utils.data import dataset_to_list

import torchvision.datasets as datasets
import torchvision.transforms as transforms
from torchvision.datasets.utils import check_integrity, download_and_extract_archive, extract_archive
import numpy as np
import PIL.Image as Image
import json
import zipfile
import os
import shutil
from tqdm import tqdm

from typing import Union
from typing import TYPE_CHECKING
import argparse    # TODO: python 3.10
if TYPE_CHECKING:
    pass


class ImageFolder(ImageSet):

    name: str = 'imagefolder'
    url: dict[str, str] = {}
    ext: dict[str, str] = {'train': '.zip', 'valid': '.zip', 'test': '.zip'}
    md5: dict[str, str] = {}
    org_folder_name: dict[str, str] = {}

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--data_format', dest='data_format', type=str,
                           help='folder, zip or numpy. (zip is using ZIP_STORED)')

    def __init__(self, data_format: str = 'folder', **kwargs):
        self.data_format: str = data_format
        if data_format == 'zip':
            raise NotImplementedError('data_format=zip Not supported! See reason at issues/42 on github.')
        super().__init__(**kwargs)
        self.param_list['imagefolder'] = ['data_format', 'url', 'org_folder_name']
        self.class_to_idx = self.get_class_to_idx()
        if self.class_to_idx is None and data_format == 'folder':
            self.class_to_idx = self.get_class_to_idx(check_folder=True)
        if self.num_classes is None:
            self.num_classes = len(self.class_to_idx)

    def middle_process(self):
        mode_list = ['valid']
        if self.data_format not in ['folder', 'zip']:
            mode_list = ['train', 'valid']
        self.data, self.targets = self.load_npz(mode_list=mode_list)

    def initialize(self, *args, **kwargs):
        if not self.check_files(data_format='folder'):
            self.initialize_folder(*args, **kwargs)
        if self.data_format == 'folder':
            self.initialize_npz(mode_list=['valid'])
        elif self.data_format == 'zip':
            self.initialize_zip(*args, **kwargs)
            self.initialize_npz(mode_list=['valid'])
        else:
            self.initialize_npz(*args, **kwargs)

    def initialize_folder(self, verbose: bool = True, img_type: str = '.jpg', **kwargs):
        print('initialize folder')
        mode_list: list[str] = ['train', 'valid'] if self.valid_set and 'valid' in self.url.keys() else ['train']
        self.class_to_idx = self.get_class_to_idx()
        idx_to_class = {v: k for k, v in self.class_to_idx.items()}
        for mode in mode_list:
            zip_path = os.path.join(self.folder_path, f'{self.name}_{mode}_store.zip')
            npz_path = os.path.join(self.folder_path, f'{self.name}_{mode}.npz')
            if os.path.isfile(zip_path):
                if verbose:
                    print('{yellow}Uncompress file{reset}: '.format(**ansi), zip_path)
                extract_archive(from_path=zip_path, to_path=self.folder_path)
                if verbose:
                    print('{green}Uncompress finished{reset}: '.format(**ansi),
                          f'{zip_path}')
                    print()
                continue
            elif os.path.isfile(npz_path):
                self.data, self.targets = self.load_npz()
                class_counters = [0] * self.num_classes
                # TODO: Parallel
                for image, target_class in self.data, self.targets:
                    image = Image.fromarray(image)
                    class_name = idx_to_class[target_class]
                    _dir = os.path.join(self.folder_path, mode, class_name)
                    if not os.path.exists(_dir):
                        os.makedirs(_dir)
                    image.save(os.path.join(_dir, f'{class_counters[target_class]}{img_type}'))
                    class_counters[target_class] += 1
                continue
            self.download_and_extract_archive(mode=mode)
            os.rename(os.path.join(self.folder_path, self.org_folder_name[mode]),
                      os.path.join(self.folder_path, mode))
            try:
                dirname = os.path.dirname(self.org_folder_name[mode])
                if dirname:
                    shutil.rmtree(os.path.join(self.folder_path, dirname))
            except FileNotFoundError:
                pass

    def download_and_extract_archive(self, mode: str):
        file_name = f'{self.name}_{mode}{self.ext[mode]}'
        file_path = os.path.normpath(os.path.join(self.folder_path, file_name))
        md5 = None if mode not in self.md5.keys() else self.md5[mode]
        if not check_integrity(file_path, md5=md5):
            prints('{yellow}Downloading Dataset{reset} '.format(**ansi),
                   f'{self.name} {mode:5s}: {file_path}', indent=10)
            download_and_extract_archive(url=self.url[mode],
                                         download_root=self.folder_path, extract_root=self.folder_path,
                                         filename=file_name, md5=md5)
            prints('{upline}{clear_line}'.format(**ansi), indent=10)
        else:
            prints('{yellow}File Already Exists{reset}: '.format(**ansi), file_path, indent=10)
            extract_archive(from_path=file_path, to_path=self.folder_path)

    def initialize_zip(self, mode_list: list[str] = ['train', 'valid'], **kwargs):
        if not self.valid_set:
            mode_list.remove('valid')
        for mode in mode_list:
            dst_path = os.path.join(self.folder_path, f'{self.name}_{mode}_store.zip')
            if not os.path.exists(dst_path):
                print('{yellow}initialize zip{reset}: '.format(**ansi), dst_path)
                src_path = os.path.normpath(os.path.join(self.folder_path, mode))
                with zipfile.ZipFile(dst_path, mode='w', compression=zipfile.ZIP_STORED) as zf:
                    for root, dirs, files in os.walk(src_path):
                        _dir = root.removeprefix(os.path.join(self.folder_path, ''))
                        for _file in files:
                            org_path = os.path.join(root, _file)
                            zip_path = os.path.join(_dir, _file)
                            zf.write(org_path, zip_path)
                print('{green}initialize zip finish{reset}'.format(**ansi))

    def initialize_npz(self, mode_list: list[str] = ['train', 'valid'],
                       transform: transforms.Lambda = transforms.Lambda(lambda x: np.array(x)),
                       **kwargs):
        if not self.valid_set:
            mode_list.remove('valid')
        json_path = os.path.join(self.folder_path, 'class_to_idx.json')
        for mode in mode_list:
            npz_path = os.path.join(self.folder_path, f'{self.name}_{mode}.npz')
            if not os.path.exists(npz_path):
                print('{yellow}initialize npz{reset}: '.format(**ansi), npz_path)
                dataset: datasets.ImageFolder = self.get_org_dataset(mode, transform=transform, data_format='folder')
                data, targets = dataset_to_list(dataset)
                data = np.stack(data)
                np.savez(npz_path, data=data, targets=targets)
                with open(json_path, 'w') as f:
                    json.dump(dataset.class_to_idx, f)
                print('{green}initialize npz finish{reset}: '.format(**ansi))

    def get_org_dataset(self, mode: str, transform: Union[str, object] = 'default',
                        data_format: str = None, **kwargs) -> Union[datasets.ImageFolder, MemoryDataset]:
        if transform == 'default':
            transform = self.get_transform(mode=mode)
        if data_format is None:
            data_format = self.data_format
            # if mode == 'valid':
            #     data_format = 'numpy'
        root = os.path.join(self.folder_path, f'{self.name}_{mode}.npz')
        if data_format == 'folder':
            root = os.path.join(self.folder_path, mode)
        elif data_format == 'zip':
            root = os.path.join(self.folder_path, f'{self.name}_{mode}_store.zip')
        DatasetClass = datasets.VisionDataset
        if data_format == 'folder':
            DatasetClass = datasets.ImageFolder
        elif data_format == 'zip':
            DatasetClass = ZipFolder
        elif not ('data' in self.__dict__.keys()):
            raise RuntimeError()   # TODO
        else:
            DatasetClass = MemoryDataset
            kwargs['data'] = self.data[mode]
            kwargs['targets'] = self.targets[mode]
        return DatasetClass(root=root, transform=transform, **kwargs)

    def load_npz(self, mode_list: list[str] = ['train', 'valid']) -> tuple[dict[str, np.ndarray], dict[str, list[int]]]:
        if not self.valid_set:
            mode_list.remove('valid')
        data = {}
        targets = {}
        for mode in mode_list:
            npz_path = os.path.join(self.folder_path, f'{self.name}_{mode}.npz')
            _dict = np.load(npz_path)
            data[mode] = _dict['data']
            targets[mode] = list(_dict['targets'])
        return data, targets

    def get_class_to_idx(self, file_path: str = None, check_folder=False) -> dict[str, int]:
        if file_path is None:
            file_path = os.path.join(self.folder_path, 'class_to_idx.json')
        if os.path.exists(file_path):
            with open(file_path) as fp:
                result: dict[str, int] = json.load(fp)
            return result
        if check_folder:
            return self.get_org_dataset('train', data_format='folder').class_to_idx
        return super().get_class_to_idx()

    def sample(self, child_name: str = None, class_dict: dict = None, sample_num: int = None, verbose=True):
        if sample_num is None:
            assert class_dict
            sample_num = len(class_dict)
        if child_name is None:
            child_name = self.name + '_sample%d' % sample_num
        src_path = self.folder_path
        mode_list = [_dir for _dir in os.listdir(
            src_path) if os.path.isdir(src_path + _dir) and _dir[0] != '.']
        dst_path = os.path.normpath(os.path.join(os.path.dirname(self.folder_path), child_name))
        if verbose:
            print('{yellow}src path{reset}: '.format(**ansi), src_path)
            print('{yellow}dst path{reset}: '.format(**ansi), dst_path)
        if class_dict is None:
            assert sample_num
            idx_list = np.arange(self.num_classes)
            np.random.seed(env['seed'])
            np.random.shuffle(idx_list)
            idx_list = idx_list[:sample_num]
            class_list = np.array(os.listdir(src_path + mode_list[0]))[idx_list]
            class_dict = {}
            for class_name in class_list:
                class_dict[class_name] = [class_name]
        if verbose:
            print(class_dict)

        len_i = len(class_dict.keys())
        for src_mode in mode_list:
            if verbose:
                print('{purple}{0}{reset}'.format(src_mode, **ansi))
            assert src_mode in ['train', 'valid', 'test', 'val']
            dst_mode = 'valid' if src_mode == 'val' else src_mode
            for i, dst_class in enumerate(class_dict.keys()):
                if not os.path.exists(os.path.join(dst_path, dst_mode, dst_class)):
                    os.makedirs(os.path.join(dst_path, dst_mode, dst_class))
                prints('{blue_light}{0}{reset}'.format(dst_class, **ansi), indent=10)
                class_list = class_dict[dst_class]
                len_j = len(class_list)
                for j, src_class in enumerate(class_list):
                    _list = os.listdir(os.path.join(src_path, src_mode, src_class))
                    prints(output_iter(i + 1, len_i) + output_iter(j + 1, len_j) +
                           f'dst: {dst_class:15s}    src: {src_class:15s}    image_num: {len(_list):>8d}', indent=10)
                    if env['tqdm']:
                        _list = tqdm(_list)
                    for _file in _list:
                        shutil.copyfile(os.path.join(src_path, src_mode, src_class, _file),
                                        os.path.join(dst_path, dst_mode, dst_class, _file))
                    if env['tqdm']:
                        print('{upline}{clear_line}'.format(**ansi))



#!/usr/bin/env python3

import os
import json

chn2eng = {
    '鱼': 'fish',
    '鸟': 'bird',
    '龟': 'turtle',
    '蜥蜴': 'lizard',
    '蛇': 'snake',
    '蜘蛛': 'spider',
    '螃蟹': 'crab',
    '狗': 'dog',
    '狼': 'wolf',
    '猫': 'cat',
    '昆虫': 'insect',
    '鼠': 'rat',
    '猴子': 'monkey',
    '菌菇': 'mushroom',
    '羊': 'sheep',
    '猪': 'pig',
    '牛': 'cow',
    '兔子': 'rabbit',
    '蝴蝶': 'butterfly',
    '狐狸': 'fox',
}
if __name__ == '__main__':

    eng2chn = {}
    for (k, v) in chn2eng.items():
        eng2chn[v] = k
    class_dict = {}
    for k in eng2chn.keys():
        class_dict[k] = []
    with open('./class.txt', encoding='utf-8') as txt:
        lines = txt.readlines()
        class_list = [a.strip().split(' ') for a in lines]
    for pair in class_list:
        print(pair)
        if pair[1] in chn2eng.keys():
            class_dict[chn2eng[pair[1]]].append(pair[0])

    with open('./class_dict.json', 'w') as f:
        json.dump(class_dict, f)



#!/usr/bin/env python3

from trojanvision.optim import PGD
from trojanzoo.defenses import Defense
import torch


class Grad_Train(Defense):

    name: str = 'grad_train'

    def __init__(self, pgd_alpha: float = 2.0 / 255, pgd_epsilon: float = 8.0 / 255, pgd_iteration: int = 7,
                 grad_lambda: float = 10, **kwargs):
        super().__init__(**kwargs)
        self.param_list['grad_train'] = ['grad_lambda']
        self.grad_lambda = grad_lambda

        self.param_list['adv_train'] = ['pgd_alpha', 'pgd_epsilon', 'pgd_iteration']
        self.pgd_alpha = pgd_alpha
        self.pgd_epsilon = pgd_epsilon
        self.pgd_iteration = pgd_iteration
        self.pgd = PGD(alpha=pgd_alpha, epsilon=pgd_epsilon, iteration=pgd_iteration, stop_threshold=None)

    def detect(self, **kwargs):
        self.model._train(loss_fn=self.loss_fn, validate_fn=self.validate_fn, verbose=True, **kwargs)

    def loss_fn(self, _input, _label, **kwargs):
        new_input = _input.repeat(4, 1, 1, 1)
        new_label = _label.repeat(4)
        noise = torch.randn_like(new_input)
        noise = noise / noise.norm(p=float('inf')) * self.pgd_epsilon
        new_input = new_input + noise
        new_input = new_input.clamp(0, 1).detach()
        new_input.requires_grad_()
        loss = self.model.loss(new_input, new_label)
        grad = torch.autograd.grad(loss, new_input, create_graph=True)[0]
        new_loss = loss + self.grad_lambda * grad.flatten(start_dim=1).norm(p=1, dim=1).mean()
        return new_loss

    def validate_fn(self, get_data_fn=None, loss_fn=None, **kwargs) -> tuple[float, float]:
        # TODO
        _, clean_acc = self.model._validate(print_prefix='Validate Clean',
                                            get_data_fn=None, **kwargs)
        _, adv_acc = self.model._validate(print_prefix='Validate Adv',
                                          get_data_fn=self.get_data, **kwargs)
        # todo: Return value
        if self.clean_acc - clean_acc > 20 and self.clean_acc > 40:
            adv_acc = 0.0
        return clean_acc, adv_acc

    def get_data(self, data: tuple[torch.Tensor, torch.Tensor], **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        _input, _label = self.model.get_data(data, **kwargs)

        def loss_fn(X: torch.FloatTensor):
            return -self.model.loss(X, _label)
        adv_x, _ = self.pgd.optimize(_input=_input, loss_fn=loss_fn)
        return adv_x, _label

    def save(self, **kwargs):
        self.model.save(folder_path=self.folder_path, suffix='_grad_train', verbose=True, **kwargs)



#!/usr/bin/env python3

from .imagemodel import _ImageModel, ImageModel
from trojanvision.datasets import ImageNet
from trojanvision.utils.bit import KNOWN_MODELS, tf2th

import torch
import torch.nn as nn
import torch.hub
import numpy as np
import os
import re
from collections import OrderedDict


class _BiT(_ImageModel):
    def __init__(self, name: str = 'BiT-M-R50x1',
                 norm_par: dict[str, list] = None,
                 **kwargs):
        norm_par = {'mean': [0.5, 0.5, 0.5],
                    'std': [0.5, 0.5, 0.5], }
        super().__init__(norm_par=norm_par, **kwargs)
        _model = KNOWN_MODELS[name](head_size=self.num_classes)
        self.features = nn.Sequential()
        OrderedDict([('root', _model.root), ('body', _model.body)])
        for name, module in _model.root.named_children():
            self.features.add_module(name=name, module=module)
        for name, module in _model.body.named_children():
            self.features.add_module(name=name, module=module)
        self.features.add_module('gn', module=getattr(_model.head, 'gn'))
        self.features.add_module('relu', module=getattr(_model.head, 'relu'))
        self.pool: nn.AdaptiveAvgPool2d = getattr(_model.head, 'avg')

        final_layer: nn.Conv2d = getattr(_model.head, 'conv')
        self.classifier = nn.Sequential(OrderedDict([
            ('fc', nn.Linear(final_layer.in_channels, final_layer.out_channels))
        ]))


class BiT(ImageModel):

    def __init__(self, name: str = 'BiT', pretrained_dataset: str = 'M',
                 layer: int = 50, width_factor: int = 1,
                 model_class: type[_BiT] = _BiT, **kwargs):
        if name[:3] == 'bit':
            name = 'BiT' + name[3:]
        full_name_list: list[str] = re.findall(r'[0-9]+|[A-Za-z]+|_', name)
        name_list = full_name_list[0].split('-')
        if len(name_list) == 1:
            name_list.append(pretrained_dataset)
            name_list.append('R')
        elif len(name_list) == 2:
            if name_list[1] == 'R':
                name_list.insert(1, pretrained_dataset)
            else:
                name_list.append('R')
        name_list[1] = name_list[1].upper()
        name_list[2] = name_list[2].upper()
        assert name_list[1] in ['S', 'M', 'L'] and name_list[2] == 'R', name
        full_name_list[0] = '-'.join(name_list)
        name = ''.join(full_name_list)

        super().__init__(name=name, layer=layer, width_factor=width_factor,
                         model_class=model_class, **kwargs)

    def get_official_weights(self, **kwargs) -> OrderedDict[str, torch.Tensor]:
        # TODO: map_location argument
        # TODO: model save_dir defaults to torch.hub.get_dir()
        file_name = f'{self.name}.npz'
        if isinstance(self.dataset, ImageNet):
            file_name = f'{self.name}-ILSVRC2012.npz'
        url = f'https://storage.googleapis.com/bit_models/{file_name}'
        print('get official model weights from: ', url)
        file_path = os.path.join(torch.hub.get_dir(), 'bit', file_name)
        if not os.path.exists(file_path):
            torch.hub.download_url_to_file(url, file_path)
        weights: dict[str, np.ndarray] = np.load(file_path)
        _dict = OrderedDict()
        _dict['features.conv.weight'] = tf2th(weights['resnet/root_block/standardized_conv2d/kernel'])
        for block_num in range(4):
            block_name = f'block{block_num+1:d}'
            block: nn.Sequential = getattr(self._model.features, block_name)
            for unit_name, unit in block.named_children():
                prefix = '/'.join(['resnet', block_name, unit_name, ''])
                dict_prefix = '.'.join(['features', block_name, unit_name, ''])
                _dict[dict_prefix + 'gn1.weight'] = tf2th(weights[prefix + 'a/group_norm/gamma'])
                _dict[dict_prefix + 'gn1.bias'] = tf2th(weights[prefix + 'a/group_norm/beta'])
                _dict[dict_prefix + 'conv1.weight'] = tf2th(weights[prefix + 'a/standardized_conv2d/kernel'])
                _dict[dict_prefix + 'gn2.weight'] = tf2th(weights[prefix + 'b/group_norm/gamma'])
                _dict[dict_prefix + 'gn2.bias'] = tf2th(weights[prefix + 'b/group_norm/beta'])
                _dict[dict_prefix + 'conv2.weight'] = tf2th(weights[prefix + 'b/standardized_conv2d/kernel'])
                _dict[dict_prefix + 'gn3.weight'] = tf2th(weights[prefix + 'c/group_norm/gamma'])
                _dict[dict_prefix + 'gn3.bias'] = tf2th(weights[prefix + 'c/group_norm/beta'])
                _dict[dict_prefix + 'conv3.weight'] = tf2th(weights[prefix + 'c/standardized_conv2d/kernel'])
                if hasattr(unit, 'downsample'):
                    weight = tf2th(weights[prefix + 'a/proj/standardized_conv2d/kernel'])
                    _dict[dict_prefix + 'downsample.weight'] = weight
        _dict['features.gn.weight'] = tf2th(weights['resnet/group_norm/gamma'])
        _dict['features.gn.bias'] = tf2th(weights['resnet/group_norm/beta'])
        _dict['classifier.fc.weight'] = tf2th(weights['resnet/head/conv2d/kernel']).flatten(1)
        _dict['classifier.fc.bias'] = tf2th(weights['resnet/head/conv2d/bias'])
        return _dict



#!/usr/bin/env python3

from .imagefolder import ImageFolder
from trojanzoo.utils.param import Module

import torchvision.transforms as transforms
import numpy as np
# import os
# import sys
# import requests
# import getpass


class VGGface2(ImageFolder):

    name: str = 'vggface2'
    data_shape = [3, 224, 224]
    num_classes = 8631
    valid_set: bool = False
    # https://docs.qingcloud.com/product/ai/deeplearning/
    url = {'train': 'https://appcenter-deeplearning.sh1a.qingstor.com/dataset/vggface2/vggface2_train.tar.gz'}
    md5 = {'train': '88813c6b15de58afc8fa75ea83361d7f'}
    org_folder_name = {'train': 'train'}

    @staticmethod
    def get_transform(mode: str) -> transforms.Compose:
        if mode == 'train':
            transform = transforms.Compose([
                transforms.RandomResizedCrop((224, 224)),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor()])
        else:
            transform = transforms.Compose([
                transforms.Resize((256, 256)),
                transforms.CenterCrop((224, 224)),
                transforms.ToTensor()])
        return transform

    def initialize_npz(self, mode_list: list[str] = ['train', 'valid'],
                       transform: transforms.Compose = transforms.Compose([transforms.Resize((256, 256)),
                                                                           transforms.Lambda(lambda x: np.array(x))]),
                       **kwargs):
        super().initialize_npz(mode_list=mode_list, transform=transform, **kwargs)

    # Out-Of-Date
    # def download(self, mode: str = 'train', file_path=None, folder_path=None, file_name=None, file_ext='.tar.gz', **kwargs):
    #     assert mode == 'train'
    #     if file_path is None:
    #         if folder_path is None:
    #             folder_path = self.folder_path
    #         if file_name is None:
    #             file_name = f'{self.name}_{mode}{file_ext}'
    #             file_path = os.path.join(folder_path, file_name)
    #     if os.path.exists(file_path['train']):
    #         print('File Already Exists: ', file_path)
    #         return file_path

    #     LOGIN_URL = "http://zeus.robots.ox.ac.uk/vgg_face2/login/"
    #     FILE_URL = "http://zeus.robots.ox.ac.uk/vgg_face2/get_file?fname=vggface2_train.tar.gz"

    #     print('Please enter your VGG Face 2 credentials:')
    #     user_string = input('    User: ')
    #     password_string = getpass.getpass(prompt='    Password: ')

    #     payload = {
    #         'username': user_string,
    #         'password': password_string
    #     }

    #     session = requests.session()
    #     r = session.get(LOGIN_URL)

    #     if 'csrftoken' in session.cookies:
    #         csrftoken = session.cookies['csrftoken']
    #     elif 'csrf' in session.cookies:
    #         csrftoken = session.cookies['csrf']
    #     else:
    #         raise ValueError("Unable to locate CSRF token.")

    #     payload['csrfmiddlewaretoken'] = csrftoken

    #     r = session.post(LOGIN_URL, data=payload)

    #     # filename = FILE_URL.split('=')[-1]

    #     with open(file_path['train'], "wb") as f:
    #         print(f"Downloading file: `{file_path['train']}`")
    #         r = session.get(FILE_URL, data=payload, stream=True)
    #         bytes_written = 0
    #         for data in r.iter_content(chunk_size=4096):
    #             f.write(data)
    #             bytes_written += len(data)
    #             MiB = bytes_written / (1024 * 1024)
    #             sys.stdout.write(f"\r{MiB:0.2f} MiB downloaded...")
    #             sys.stdout.flush()

    #     print("\nDone.")
    #     return file_path


class Sample_VGGface2(VGGface2):

    name: str = 'sample_vggface2'
    num_classes = 20
    url = {}
    org_folder_name = {}

    def initialize_folder(self):
        _dict = Module(self.__dict__)
        _dict.__delattr__('folder_path')
        vggface2 = VGGface2(**_dict)
        vggface2.sample(child_name=self.name, sample_num=self.num_classes)



#!/usr/bin/env python3

from .fonts import *
from trojanzoo.utils import to_numpy

import os
import numpy as np
import torch

import matplotlib.ticker as ticker
from matplotlib import rc
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from matplotlib.lines import Line2D
from matplotlib.container import BarContainer
import seaborn
import scipy.stats as stats

from scipy.interpolate import UnivariateSpline
# from scipy.optimize import curve_fit

rc('font', family='serif', serif='Palatino', weight='bold')
rc('svg', image_inline=True, fonttype='none')
rc('pdf', fonttype=42)
rc('ps', fonttype=42)
rc('mathtext', fontset='cm')


class Figure:
    def __init__(self, name: str, folder_path: str = None, fig: Figure = None, ax: Axes = None, figsize: tuple[float, float] = (5, 2.5), tex=False):
        super(Figure, self).__init__()
        if tex:
            rc('text', usetex=True)
        self.name: str = name
        self.folder_path: str = folder_path
        if folder_path is None:
            self.folder_path = './output/'
        self.fig: Figure = fig
        self.ax: Axes = ax
        if fig is None and ax is None:
            self.fig: Figure = plt.figure(figsize=figsize)
            self.ax = self.fig.add_subplot(1, 1, 1)
        self.ax.spines['top'].set_visible(False)
        self.ax.spines['bottom'].set_visible(True)
        self.ax.spines['left'].set_visible(False)
        self.ax.spines['right'].set_visible(False)
        self.ax.grid(axis='y', linewidth=1)
        self.ax.set_axisbelow(True)

        self.ax.set_xlim([0.0, 1.0])
        self.ax.set_ylim([0.0, 1.0])

    def set_legend(self, *args, frameon: bool = True, edgecolor='white', framealpha=1.0,
                   fontsize=11, fontstyle='italic', **kwargs) -> None:
        self.ax.legend(*args, frameon=frameon, edgecolor=edgecolor, framealpha=framealpha, **kwargs)
        plt.setp(self.ax.get_legend().get_texts(), fontsize=fontsize, fontstyle=fontstyle)

    def set_axis_label(self, axis: str, text: str, fontsize: int = 12,
                       family='serif', font='Palatino', weight='bold', **kwargs):
        getattr(self.ax, f'set_{axis}label')(text, fontsize=fontsize,
                                             family=family, font=font, weight=weight, **kwargs)

    def set_title(self, text: str = None, fontsize: int = 16, fontweight: str = 'bold') -> None:
        if text is None:
            text = self.name
        self.ax.set_title(text, fontsize=fontsize, fontweight=fontweight)

    def save(self, path: str = None, folder_path: str = None, ext: str = '.pdf') -> None:
        if path is None:
            if folder_path is None:
                folder_path = self.folder_path
            path = os.path.join(folder_path, f'{self.name}{ext}')
        else:
            folder_path = os.path.dirname(path)
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
        self.fig.savefig(path, dpi=100, bbox_inches='tight')

    def set_axis_lim(self, axis: str, lim: list[float] = [0.0, 1.0], margin: list[float] = [0.0, 0.0],
                     piece: int = 10, _format: str = '%.1f',
                     fontsize: int = 11) -> None:
        if _format == 'integer':
            _format = '%d'
        lim_func = getattr(self.ax, f'set_{axis}lim')
        set_ticks_func = getattr(self.ax, f'set_{axis}ticks')

        def format_func(_str):
            getattr(self.ax, f'{axis}axis').set_major_formatter(
                ticker.FormatStrFormatter(_str))
        ticks = np.append(
            np.arange(lim[0], lim[1], (lim[1] - lim[0]) / piece), lim[1])
        final_lim = [lim[0] - margin[0], lim[1] + margin[1]]
        lim_func(final_lim)
        set_ticks_func(ticks)
        ticks = getattr(self.ax, f'get_{axis}ticks')()
        set_ticklabels_func = getattr(self.ax, f'set_{axis}ticklabels')
        set_ticklabels_func(ticks, fontsize=fontsize)
        format_func(_format)

    def curve(self, x: np.ndarray, y: np.ndarray, color: str = 'black', linewidth: int = 2,
              label: str = None, markerfacecolor: str = 'white', linestyle: str = '-', zorder: int = 1, **kwargs) -> Line2D:
        # linestyle marker markeredgecolor markeredgewidth markerfacecolor markersize alpha
        ax = seaborn.lineplot(x, y, ax=self.ax, color=color, linewidth=linewidth,
                              markerfacecolor=markerfacecolor, zorder=zorder, **kwargs)
        line: Line2D = ax.get_lines()[-1]
        line.set_linestyle(linestyle)
        if label is not None:
            self.curve_legend(label=label, color=color, linewidth=linewidth, **kwargs)
        return line

    def curve_legend(self, label: str = None, color: str = 'black', linewidth: int = 2, markerfacecolor: str = 'white', **kwargs) -> Line2D:
        # linestyle marker markeredgecolor markeredgewidth markerfacecolor markersize alpha
        line, = self.ax.plot([], [], color=color, linewidth=linewidth, markeredgewidth=linewidth, markeredgecolor=color,
                             label=label, markerfacecolor=markerfacecolor, **kwargs)
        return line

    def scatter(self, x: np.ndarray, y: np.ndarray, color: str = 'black', linewidth: int = 2,
                label: str = None, marker: str = 'D', facecolor: str = 'white', zorder: int = 3, **kwargs):
        # marker markeredgecolor markeredgewidth markerfacecolor markersize alpha
        if label is not None:
            self.curve_legend(label=label, color=color, linewidth=linewidth, marker=marker, **kwargs)
        return self.ax.scatter(x, y, color=color, linewidth=linewidth, marker=marker, facecolor=facecolor, zorder=zorder, **kwargs)

# Markers
# '.' point marker
# ',' pixel marker
# 'o' circle marker
# 'v' triangle_down marker
# '^' triangle_up marker
# '<' triangle_left marker
# '>' triangle_right marker
# '1' tri_down marker
# '2' tri_up marker
# '3' tri_left marker
# '4' tri_right marker
# 's' square marker
# 'p' pentagon marker
# '*' star marker
# 'h' hexagon1 marker
# 'H' hexagon2 marker
# '+' plus marker
# 'x' x marker
# 'D' diamond marker
# 'd' thin_diamond marker
# '|' vline marker
# '_' hline marker

# Line Styles
# '-'     solid line style
# '--'    dashed line style
# '-.'    dash-dot line style
# ':'     dotted line style

    def bar(self, x: np.ndarray, y: np.ndarray, color: str = 'black', width: float = 0.2,
            align: str = 'edge', edgecolor: str = 'white', label: str = None, **kwargs) -> BarContainer:
        # facecolor edgewidth alpha
        return self.ax.bar(x, y, color=color, width=width, align=align, edgecolor=edgecolor, label=label, **kwargs)

    def bar3d(self, x: np.ndarray, y: np.ndarray, z: np.array, color: str = 'black', size: tuple[float, float] = 0.5,
              label: str = None, **kwargs) -> BarContainer:
        # facecolor edgewidth alpha
        if isinstance(size, (float, int)):
            size = [size, size]
        return self.ax.bar3d(x=x, y=y, z=np.zeros_like(x),
                             dx=np.ones_like(x) * size[0], dy=np.ones_like(y) * size[1], dz=z,
                             color=color, label=label, **kwargs)

    def hist(self, x: np.ndarray, bins: list[float] = None, normed: bool = True, **kwargs):
        return self.ax.hist(x, bins=bins, normed=normed, **kwargs)

    def autolabel(self, rects: BarContainer, above: bool = True,
                  fontsize: int = 6):
        """Attach a text label above each bar in *rects*, displaying its height."""
        for rect in rects:
            height = int(rect.get_height())
            offset = 3 if above else -13
            self.ax.annotate('%d' % (abs(height)),
                             xy=(rect.get_x() + rect.get_width() / 2, height),
                             xytext=(0, offset),  # 3 points vertical offset
                             textcoords="offset points",
                             ha='center', va='bottom', fontsize=fontsize)

    @staticmethod
    def get_roc_curve(label: list[int], pred: list[int]) -> tuple[list[float], list[float]]:
        total_inst = len(label)
        total_pos_inst = len(np.where(label == 1)[0])

        assert len(label) == len(pred)
        # true positive rates and false positive rates
        tprs, fprs, thresholds = [], [], []

        # iterate over all positive thresholds
        for threshold in np.unique(pred):
            pred_pos_idx: np.ndarray = np.where(pred >= threshold)[0]
            # number of predicted positive instances
            pred_pos_inst = len(pred_pos_idx)
            # number of true positive instances
            true_pos_inst: int = np.count_nonzero(label[pred_pos_idx])

            tpr = true_pos_inst / total_pos_inst
            fpr = (pred_pos_inst - true_pos_inst) / (total_inst - total_pos_inst)
            tprs.append(tpr)
            fprs.append(fpr)
            thresholds.append(threshold)
        return fprs, tprs

    @staticmethod
    def normalize(x: np.ndarray, _min: float = None, _max: float = None, tgt_min: float = 0.0, tgt_max: float = 1.0) -> np.ndarray:
        x = to_numpy(x)
        if _min is None:
            _min = x.min()
        if _max is None:
            _max = x.max()
        x = (x - _min) / (_max - _min) * (tgt_max - tgt_min) + tgt_min
        return x

    @staticmethod
    def groups_err_bar(x: np.ndarray, y: np.ndarray) -> dict[float, np.ndarray]:
        y_dict = {}
        for _x in set(x):
            y_dict[_x] = np.array([y[t] for t in range(len(y)) if x[t] == _x])
        return y_dict

    @staticmethod
    def flatten_err_bar(y_dict: dict[float, np.ndarray]) -> tuple[np.ndarray, np.ndarray]:
        x = []
        y = []
        for _x in y_dict.keys():
            for _y in y_dict[_x]:
                x.append(_x)
                y.append(_y)
        return np.array(x), np.array(y)

    @classmethod
    def normalize_err_bar(cls, x: np.ndarray, y: np.ndarray):
        x = cls.normalize(x)
        y_dict = cls.groups_err_bar(x, y)
        y_mean = np.array([y_dict[_x].mean()
                           for _x in np.sort(list(y_dict.keys()))])
        y_norm = cls.normalize(y_mean)
        y_dict = cls.adjust_err_bar(y_dict, y_norm - y_mean)
        return cls.flatten_err_bar(y_dict)

    @classmethod
    def avg_smooth_err_bar(cls, x: np.ndarray, y: np.ndarray, window: int = 3):
        y_dict = cls.groups_err_bar(x, y)
        y_mean = np.array([y_dict[_x].mean()
                           for _x in np.sort(list(y_dict.keys()))])
        y_smooth = cls.avg_smooth(y_mean, window=window)
        y_dict = cls.adjust_err_bar(y_dict, y_smooth - y_mean)
        return cls.flatten_err_bar(y_dict)

    @staticmethod
    def adjust_err_bar(y_dict: dict[float, np.ndarray], mean: np.ndarray = None, std: np.ndarray = None) -> dict[float, np.ndarray]:
        sort_keys = np.sort(list(y_dict.keys()))
        if isinstance(mean, float):
            mean = mean * np.ones(len(sort_keys))
        if isinstance(std, float):
            std = std * np.ones(len(sort_keys))
        for i in range(len(sort_keys)):
            key = sort_keys[i]
            if mean:
                y_dict[key] = y_dict[key] + mean[i]
            if std:
                y_dict[key] = y_dict[key].mean() + \
                    (y_dict[key] - y_dict[key].mean()) * std[i]
        return y_dict

    @staticmethod
    def avg_smooth(x: np.ndarray, window: int = 3) -> np.ndarray:
        _x = torch.as_tensor(x)
        new_x = torch.zeros_like(_x)
        for i in range(len(_x)):
            if i < window // 2:
                new_x[i] = (_x[0] * (window // 2 - i)
                            + _x[: i + (window + 1) // 2].sum()) / window
            elif i >= len(_x) - (window - 1) // 2:
                new_x[i] = (_x[-1] * ((window + 1) // 2 - len(_x) + i) +
                            _x[i - window // 2:].sum()) / window
            else:
                new_x[i] = _x[i - window // 2:i + 1 + (window - 1) // 2].mean()
        return to_numpy(new_x) if isinstance(x, np.ndarray) else new_x

    @staticmethod
    def poly_fit(x: np.ndarray, y: np.ndarray, x_grid: np.ndarray, degree: int = 1) -> np.ndarray:
        fit_data = to_numpy(y)
        z = np.polyfit(x, fit_data, degree)
        y_grid = np.polyval(z, x_grid)
        return y_grid

    @staticmethod
    def tanh_fit(x: np.ndarray, y: np.ndarray, x_grid: np.ndarray,
                 degree: int = 1, mean_bias: float = 0.0, scale_multiplier: float = 1.0) -> np.ndarray:
        mean = (max(y) + min(y)) / 2 + mean_bias
        scale = max(abs(y - mean)) * scale_multiplier
        fit_data = to_numpy(torch.as_tensor((y - mean) / scale).atanh())
        z = np.polyfit(x, fit_data, degree)
        y_grid = np.tanh(np.polyval(z, x_grid)) * scale + mean
        return y_grid

    @staticmethod
    def atan_fit(x: np.ndarray, y: np.ndarray, x_grid: np.ndarray,
                 degree: int = 1, mean_bias: float = 0.0, scale_multiplier: float = 1.0) -> np.ndarray:
        mean = (max(y) + min(y)) / 2 + mean_bias
        scale = max(abs(y - mean)) * scale_multiplier
        fit_data = to_numpy(torch.as_tensor((y - mean) / scale).tan())
        z = np.polyfit(x, fit_data, degree)
        y_grid = np.tanh(np.polyval(z, x_grid)) * scale + mean
        return y_grid

    @staticmethod
    def exp_fit(x: np.ndarray, y: np.ndarray, x_grid: np.ndarray,
                degree: int = 1, increase: bool = True, epsilon: float = 0.01) -> np.ndarray:
        y_max = max(y)
        y_min = min(y)
        if increase:
            fit_data = np.log(y + epsilon - y_min)
        else:
            fit_data = np.log(y_max + epsilon - y)

        z = np.polyfit(x, fit_data, degree)
        y_grid = np.exp(np.polyval(z, x_grid))
        if increase:
            y_grid += y_min - epsilon
        else:
            y_grid = y_max + epsilon - y_grid
        return y_grid

    @staticmethod
    def inverse_fit(x: np.ndarray, y: np.ndarray, x_grid: np.ndarray,
                    degree: int = 1, y_lower_bound: float = 0.0) -> np.ndarray:
        fit_data = 1 / (y - y_lower_bound)
        z = np.polyfit(x, fit_data, degree)
        y_grid = 1 / (np.polyval(z, x_grid)) + y_lower_bound
        return y_grid

    @staticmethod
    def monotone(x: np.ndarray, increase: bool = True) -> np.ndarray:
        temp = 0.0
        y = np.copy(x)
        if increase:
            temp = min(x)
        else:
            temp = max(x)
        for i in range(len(x)):
            if ((increase and x[i] < temp) or (not increase and x[i] > temp)):
                y[i] = temp
            else:
                temp = x[i]
        return y

    @staticmethod
    def gaussian_kde(x: np.ndarray, x_grid: np.ndarray) -> np.ndarray:
        kde_func = stats.gaussian_kde(x)
        y_grid = kde_func(x_grid)
        return y_grid

    @staticmethod
    def interp_fit(x: np.ndarray, y: np.ndarray, x_grid: np.ndarray, interp_num: int = 20) -> np.ndarray:
        func = UnivariateSpline(x, y, s=interp_num)
        y_grid = func(x_grid)
        return y_grid



#!/usr/bin/env python3

from .adv import *
from .poison import *
from .backdoor import *

from trojanvision.configs import Config, config
from trojanvision.datasets import ImageSet
import trojanzoo.attacks
from trojanzoo.attacks import Attack
import argparse
from typing import Union

class_dict = {
    # adversarial attack
    'pgd': PGD,
    # 'advmind': AdvMind,

    # poisoning attack
    'poison_basic': PoisonBasic,
    'imc_poison': IMC_Poison,

    # backdoor attack
    'badnet': BadNet,
    'trojannn': TrojanNN,
    'latent_backdoor': LatentBackdoor,
    'imc': IMC,
    'reflection_backdoor': ReflectionBackdoor,
    'bypass_embed': BypassEmbed,
    'trojannet': TrojanNet,
    'clean_label': CleanLabel,
    'hidden_trigger': HiddenTrigger,

    'term_study': TermStudy,
    'unlearn': Unlearn,

    # imc adaptive settings
    'imc_latent': IMC_Latent,
    'imc_advtrain': IMC_AdvTrain,
    'imc_strip': IMC_STRIP,
    'imc_multi': IMC_Multi,
    'imc_magnet': IMC_MagNet,
    'imc_abs': IMC_ABS,
    'imc_adaptive': IMC_Adaptive,
}


def add_argument(parser: argparse.ArgumentParser, attack_name: str = None, attack: Union[str, Attack] = None,
                 class_dict: dict[str, type[Attack]] = class_dict):
    return trojanzoo.attacks.add_argument(parser=parser, attack_name=attack_name, attack=attack,
                                          class_dict=class_dict)


def create(attack_name: str = None, attack: Union[str, Attack] = None,
           dataset_name: str = None, dataset: Union[str, ImageSet] = None,
           config: Config = config, class_dict: dict[str, type[Attack]] = class_dict, **kwargs):
    return trojanzoo.attacks.create(attack_name=attack_name, attack=attack,
                                    dataset_name=dataset_name, dataset=dataset,
                                    config=config, class_dict=class_dict, **kwargs)



#!/usr/bin/env python3
from .badnet import BadNet
from .trojannet_utils import MLPNet, Combined_Model
from trojanvision.environ import env
from trojanzoo.utils.output import prints

import torch
import numpy as np

import os
from itertools import combinations
from scipy.special import comb
import argparse
from typing import Callable


class TrojanNet(BadNet):
    name: str = "trojannet"

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--select_point', dest='select_point', type=int,
                           help='the number of select_point, defaults to 2')

    def __init__(self, select_point: int = 2, **kwargs):
        super().__init__(**kwargs)
        self.param_list['trojannet'] = ['select_point', 'mlp_dim']
        self.all_point = self.mark.mark_height * self.mark.mark_width
        self.select_point = select_point

        self.x, self.y = self.synthesize_training_sample()
        self.mark.org_mark = self.x[self.target_class].repeat(
            self.dataset.data_shape[0], 1).view(self.mark.org_mark.shape)
        self.mark.mark, _, _ = self.mark.mask_mark(height_offset=self.mark.height_offset,
                                                   width_offset=self.mark.width_offset)
        self.mlp_dim = len(self.y) + 1
        self.mlp_model = MLPNet(input_dim=self.all_point, output_dim=self.mlp_dim,
                                dataset=self.dataset, loss_weights=None)
        self.combined_model = Combined_Model(org_model=self.model._model, mlp_model=self.mlp_model._model,
                                             mark=self.mark, dataset=self.dataset)

    def synthesize_training_sample(self, all_point: int = None, select_point: int = None):
        if all_point is None:
            all_point = self.all_point
        if select_point is None:
            select_point = self.select_point
        if 2**all_point < self.model.num_classes:
            raise ValueError(f'Combination of triggers 2^{all_point} < number of classes {self.model.num_classes} !')
        combination_list = []
        for i in range(all_point):
            if len(combination_list) >= self.model.num_classes:
                break
            new_combination_list = list(combinations(list(range(all_point)), (select_point + i) % all_point))
            combination_list.extend(new_combination_list)
        np.random.seed(env['seed'])
        np.random.shuffle(combination_list)

        x = torch.ones(len(combination_list), all_point, dtype=torch.float)
        for i, idx in enumerate(combination_list):
            x[i][list(idx)] = 0.0
        y = list(range(len(combination_list)))
        return x, y

    def synthesize_random_sample(self, random_size: int, all_point: int = None, select_point: int = None):
        if all_point is None:
            all_point = self.all_point
        if select_point is None:
            select_point = self.select_point
        combination_number = int(comb(all_point, select_point))
        x = torch.rand(random_size, all_point) + 2 * torch.rand(1) - 1
        x = x.clamp(0, 1)
        y = [combination_number] * random_size
        return x, y

    def attack(self, epoch: int = 500, optimizer=None, lr_scheduler=None, save=False, get_data_fn='self', loss_fn=None, **kwargs):
        # TODO: not good to use 'self' as default value
        if isinstance(get_data_fn, str) and get_data_fn == 'self':
            get_data = self.get_data
        if isinstance(loss_fn, str) and loss_fn == 'self':
            loss_fn = self.loss_fn
        train_x, train_y = self.x, self.y
        valid_x, valid_y = self.x, self.y
        loader_train = [(train_x, torch.tensor(train_y, dtype=torch.long))]
        loader_valid = [(valid_x, torch.tensor(valid_y, dtype=torch.long))]

        optimizer = torch.optim.Adam(params=self.mlp_model.parameters(), lr=1e-2)
        self.mlp_model._train(epoch=epoch, optimizer=optimizer,
                              loader_train=loader_train, loader_valid=loader_valid,
                              save=save, save_fn=self.save)
        self.validate_fn()

    def save(self, **kwargs):
        filename = self.get_filename(**kwargs)
        file_path = os.path.join(self.folder_path, filename)
        self.mlp_model.save(file_path + '.pth', verbose=True)

    def load(self, **kwargs):
        filename = self.get_filename(**kwargs)
        file_path = os.path.join(self.folder_path, filename)
        self.mlp_model.load(file_path + '.pth', verbose=True)

    def validate_fn(self,
                    get_data_fn: Callable[..., tuple[torch.Tensor, torch.Tensor]] = None,
                    loss_fn: Callable[..., torch.Tensor] = None,
                    main_tag: str = 'valid', indent: int = 0, **kwargs) -> tuple[float, float]:
        _, clean_acc = self.combined_model._validate(print_prefix='Validate Clean', main_tag='valid clean',
                                                     get_data_fn=None, indent=indent, **kwargs)
        _, target_acc = self.combined_model._validate(print_prefix='Validate Trigger Tgt', main_tag='valid trigger target',
                                                      get_data_fn=self.get_data, keep_org=False, poison_label=True,
                                                      indent=indent, **kwargs)
        self.combined_model._validate(print_prefix='Validate Trigger Org', main_tag='',
                                      get_data_fn=self.get_data, keep_org=False, poison_label=False,
                                      indent=indent, **kwargs)
        prints(f'Validate Confidence: {self.validate_confidence():.3f}', indent=indent)
        prints(f'Neuron Jaccard Idx: {self.check_neuron_jaccard():.3f}', indent=indent)
        if self.clean_acc - clean_acc > 3 and self.clean_acc > 40:  # TODO: better not hardcoded
            target_acc = 0.0
        return clean_acc, target_acc



#!/usr/bin/env python3

from .imagefolder import ImageFolder
from trojanzoo.utils.output import ansi, prints

import torchvision.transforms as transforms
from torchvision.datasets.utils import download_file_from_google_drive, extract_archive, check_integrity
import numpy as np
import os
import shutil
import pandas as pd

from trojanvision import __file__ as root_file
root_dir = os.path.dirname(root_file)


class CUB200(ImageFolder):

    name = 'cub200'
    data_shape = [3, 224, 224]
    num_classes = 200
    valid_set = True
    # http://www.vision.caltech.edu/visipedia/CUB-200.html
    # http://www.vision.caltech.edu/visipedia-data/CUB-200/images.tgz
    url = {'train': '1GDr1OkoXdhaXWGA8S3MAq3a522Tak-nx'}
    ext = {'train': '.tgz'}
    md5 = {'train': '2bbe304ef1aa3ddb6094aa8f53487cf2'}

    org_folder_name = {'train': 'images'}

    @staticmethod
    def get_transform(mode: str) -> transforms.Compose:
        if mode == 'train':
            transform = transforms.Compose([
                transforms.RandomResizedCrop((224, 224)),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor()])
        else:
            transform = transforms.Compose([
                transforms.Resize((256, 256)),
                transforms.CenterCrop((224, 224)),
                transforms.ToTensor()])
            # BiT transform
            # transform = transforms.Compose([
            #     transforms.Resize((480, 480)),
            #     transforms.ToTensor()])
        return transform

    def initialize_npz(self, mode_list: list[str] = ['train', 'valid'],
                       transform: transforms.Compose = transforms.Compose([transforms.Resize((256, 256)),
                                                                           transforms.Lambda(lambda x: np.array(x))]),
                       **kwargs):
        super().initialize_npz(mode_list=mode_list, transform=transform, **kwargs)

    def download_and_extract_archive(self, mode: str):
        file_name = f'{self.name}_{mode}{self.ext[mode]}'
        file_path = os.path.normpath(os.path.join(self.folder_path, file_name))
        md5 = None if mode not in self.md5.keys() else self.md5[mode]
        if not check_integrity(file_path, md5=md5):
            prints('{yellow}Downloading Dataset{reset} '.format(**ansi),
                   f'{self.name} {mode:5s}: {file_path}', indent=10)
            download_file_from_google_drive(file_id=self.url[mode],
                                            root=self.folder_path, filename=file_name, md5=md5)
            print('{upline}{clear_line}'.format(**ansi))
        else:
            prints('{yellow}File Already Exists{reset}: '.format(**ansi), file_path, indent=10)
        extract_archive(from_path=file_path, to_path=self.folder_path)

    def initialize_folder(self, verbose: bool = True, img_type: str = '.jpg', **kwargs):
        super().initialize_folder(verbose=verbose, img_type=img_type, **kwargs)
        # Remove useless files
        # os.remove(os.path.join(self.folder_path, ''))
        # dirpath = os.path.join(self.folder_path, 'train')
        # for fpath in os.listdir(dirpath):
        #     path = os.path.join(dirpath, fpath)
        #     if os.path.isfile(path):
        #         os.remove(path)

        # Split Train and Valid Set
        txt_path = os.path.normpath(os.path.join(root_dir, 'data', self.name, 'test.txt'))
        file_list: list[str] = []
        with open(txt_path, 'r') as fp:
            file_list = fp.read().split('\n')[:-1]
        src_dir = os.path.join(self.folder_path, 'train')
        dst_dir = os.path.join(self.folder_path, 'valid')
        if not os.path.exists(dst_dir):
            os.makedirs(dst_dir)
        for fpath in file_list:
            src_path = os.path.join(src_dir, fpath)
            dst_path = os.path.join(dst_dir, fpath)
            dir_name = os.path.dirname(dst_path)
            if not os.path.exists(dir_name):
                os.makedirs(dir_name)
            shutil.move(src_path, dst_path)


class CUB200_2011(CUB200):

    name = 'cub200_2011'
    # http://www.vision.caltech.edu/visipedia/CUB-200-2011.html
    # http://www.vision.caltech.edu/visipedia-data/CUB-200-2011/CUB_200_2011.tgz
    url = {'train': '1hbzc_P1FuxMkcabkgn9ZKinBwW683j45'}
    ext = {'train': '.tgz'}
    md5 = {'train': '97eceeb196236b17998738112f37df78'}

    org_folder_name = {'train': 'CUB_200_2011/images'}

    def initialize_folder(self, verbose: bool = True, img_type: str = '.jpg', **kwargs):
        super(CUB200, self).initialize_folder(verbose=verbose, img_type=img_type, **kwargs)
        # Split Train and Valid Set
        src_dir = os.path.join(self.folder_path, 'total')
        dst_dir = {'train': os.path.join(self.folder_path, 'train'),
                   'valid': os.path.join(self.folder_path, 'valid')}
        os.rename(dst_dir['train'], src_dir)
        os.remove(os.path.join(self.folder_path, 'attributes.txt'))

        images = pd.read_csv(os.path.join(root_dir, 'data', self.name, 'images.txt'),
                             sep=' ', names=['img_id', 'filepath'])
        train_test_split = pd.read_csv(os.path.join(root_dir, 'data', self.name, 'train_test_split.txt'),
                                       sep=' ', names=['img_id', 'is_training_img'])
        data = images.merge(train_test_split, on='img_id')
        file_dict: dict[str, list[str]] = {
            'train': data[data.is_training_img == 1]['filepath'].tolist(),
            'valid': data[data.is_training_img == 0]['filepath'].tolist(),
        }
        for mode in ['train', 'valid']:
            for fpath in file_dict[mode]:
                src_path = os.path.join(src_dir, fpath)
                dst_path = os.path.join(dst_dir[mode], fpath)
                dir_name = os.path.dirname(dst_path)
                if not os.path.exists(dir_name):
                    os.makedirs(dir_name)
                shutil.move(src_path, dst_path)
        shutil.rmtree(src_dir)



#!/usr/bin/env python3

from .version import __version__

import trojanzoo.environ
import trojanzoo.datasets
import trojanzoo.models
import trojanzoo.trainer
from trojanzoo.utils.tensor import to_tensor, to_numpy, to_list

__all__ = ['to_tensor', 'to_numpy', 'to_list']

# import trojanzoo.utils
# import trojanzoo.configs
# import trojanzoo.optim



#!/usr/bin/env python3


from trojanzoo.models import Model
from trojanzoo.configs import config
from trojanzoo.utils import get_name
from trojanzoo.utils.output import ansi, prints

from torch.utils.tensorboard import SummaryWriter

from typing import TYPE_CHECKING
from trojanzoo.configs import Config    # TODO: python 3.10
from trojanzoo.datasets import Dataset
from torch.optim.optimizer import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
import argparse
if TYPE_CHECKING:
    pass


class Trainer:
    name = 'trainer'
    param_list: list[str] = ['optim_args', 'train_args', 'writer_args',
                             'optimizer', 'lr_scheduler', 'writer']

    @staticmethod
    def add_argument(group: argparse._ArgumentGroup) -> argparse._ArgumentGroup:
        group.add_argument('--epoch', dest='epoch', type=int,
                           help='training epochs, defaults to config[train][epoch].')
        group.add_argument('--lr', dest='lr', type=float,
                           help='learning rate, defaults to 0.1.')
        group.add_argument('--parameters', dest='parameters', default='full',
                           help='training parameters (\'features\', \'classifier\', \'full\'), defaults to \'full\'.')
        group.add_argument('--OptimType', dest='OptimType',
                           help='optimizer type, defaults to SGD.')
        group.add_argument('--momentum', dest='momentum', type=float,
                           help='momentum passed to Optimizer, defaults to 0.9.')
        group.add_argument('--weight_decay', dest='weight_decay', type=float,
                           help='momentum passed to Optimizer, defaults to 1e-4.')
        group.add_argument('--nesterov', dest='nesterov', action='store_true',
                           help='use nesterov for SGD.')
        group.add_argument('--lr_scheduler', dest='lr_scheduler', action='store_true',
                           help='use torch.optim.lr_scheduler.StepLR.')
        group.add_argument('--lr_step_size', dest='lr_step_size', type=int,
                           help='step_size passed to torch.optim.lr_scheduler.StepLR, defaults to 50.')
        group.add_argument('--lr_gamma', dest='lr_gamma', type=float,
                           help='gamma passed to torch.optim.lr_scheduler.StepLR, defaults to 0.1.')
        group.add_argument('--amp', dest='amp', action='store_true',
                           help='Automatic Mixed Precision.')
        group.add_argument('--validate_interval', dest='validate_interval', type=int,
                           help='validate interval during training epochs, defaults to 10.')
        group.add_argument('--save', dest='save', action='store_true',
                           help='save training results.')
        group.add_argument('--tensorboard', dest='tensorboard', action='store_true',
                           help='save training logging for tensorboard.')
        group.add_argument('--log_dir', dest='log_dir',
                           help='save training logging for tensorboard.')
        group.add_argument('--flush_secs', dest='flush_secs', type=int,
                           help='How often, in seconds, to flush the pending events and summaries to disk.')
        return group

    def __init__(self, optim_args: dict = {}, train_args: dict = {}, writer_args: dict = {},
                 optimizer: Optimizer = None, lr_scheduler: _LRScheduler = None,
                 writer: SummaryWriter = None):
        self.optim_args = {} | optim_args   # to avoid BadAppend issues
        self.train_args = {} | train_args
        self.writer_args = {} | writer_args
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler
        self.writer = writer

    def __getitem__(self, key: str):
        if key in self.train_args.keys():
            return self.train_args[key]
        return getattr(self, key)

    def __getattr__(self, key: str):
        if key in self.train_args.keys():
            return self.train_args[key]
        raise AttributeError(key)

    def keys(self) -> list[str]:
        keys: list[str] = self.param_list.copy()
        keys.remove('optim_args')
        keys.remove('train_args')
        keys.remove('writer_args')
        keys.extend(list(self.train_args.keys()))
        return keys

    def summary(self, indent: int = 0):
        prints('{blue_light}{0:<20s}{reset} Parameters: '.format(self.name, **ansi), indent=indent)
        for key in self.param_list:
            value = getattr(self, key)
            if value is not None:
                prints('{green}{0:<10s}{reset}'.format(key, **ansi), indent=indent + 10)
                prints(value, indent=indent + 10)
                prints('-' * 20, indent=indent + 10)


def add_argument(parser: argparse.ArgumentParser, ClassType: type[Trainer] = Trainer) -> argparse._ArgumentGroup:
    group = parser.add_argument_group('{yellow}trainer{reset}'.format(**ansi))
    return ClassType.add_argument(group)


def create(dataset_name: str = None, dataset: Dataset = None, model: Model = None,
           ClassType: type[Trainer] = Trainer,
           tensorboard: bool = None, config: Config = config, **kwargs) -> tuple[Optimizer, _LRScheduler, dict]:
    assert isinstance(model, Model)
    dataset_name = get_name(name=dataset_name, module=dataset, arg_list=['-d', '--dataset'])
    result = config.get_config(dataset_name=dataset_name)['trainer']._update(kwargs)

    optim_keys = model.define_optimizer.__code__.co_varnames
    train_keys = model._train.__code__.co_varnames
    writer_keys = SummaryWriter.__init__.__code__.co_varnames   # log_dir, flush_secs, ...
    optim_args = {}
    train_args = {}
    writer_args = {}
    for key, value in result.items():
        if key in optim_keys:
            _dict = optim_args
        elif key in train_keys:
            _dict = train_args
        elif key in writer_keys:
            _dict = writer_args
        else:
            continue
        _dict[key] = value
    optimizer, lr_scheduler = model.define_optimizer(**optim_args)
    writer = SummaryWriter(**writer_args) if tensorboard else None
    return ClassType(optim_args=optim_args, train_args=train_args,
                     optimizer=optimizer, lr_scheduler=lr_scheduler,
                     writer=writer)



#!/usr/bin/env python3

from trojanvision.models.imagemodel import ImageModel, _ImageModel
from trojanvision.marks import Watermark

import torch
import torch.nn as nn


class _MLPNet(nn.Module):
    def __init__(self, input_dim: int, output_dim: int, **kwargs):
        super().__init__()
        self.ly1 = nn.Linear(in_features=input_dim, out_features=8)
        self.relu1 = nn.ReLU()
        self.ly1_bn = nn.BatchNorm1d(num_features=8)
        self.ly2 = nn.Linear(in_features=8, out_features=8)
        self.relu2 = nn.ReLU()
        self.ly2_bn = nn.BatchNorm1d(num_features=8)
        self.ly3 = nn.Linear(in_features=8, out_features=8)
        self.relu3 = nn.ReLU()
        self.ly3_bn = nn.BatchNorm1d(num_features=8)
        self.ly4 = nn.Linear(in_features=8, out_features=8)
        self.relu4 = nn.ReLU()
        self.ly4_bn = nn.BatchNorm1d(num_features=8)
        self.output = nn.Linear(in_features=8, out_features=output_dim)

    def forward(self, x, **kwargs):
        x = self.ly1_bn(self.relu1(self.ly1(x)))
        x = self.ly2_bn(self.relu2(self.ly2(x)))
        x = self.ly3_bn(self.relu3(self.ly3(x)))
        x = self.ly4_bn(self.relu4(self.ly4(x)))
        x = self.output(x)
        return x


class MLPNet(ImageModel):
    def __init__(self, name='mlpnet', model_class=_MLPNet, **kwargs):
        super().__init__(name=name, model_class=model_class, **kwargs)

    def get_logits(self, _input: torch.Tensor, **kwargs):
        return self._model(_input, **kwargs)


class _Combined_Model(_ImageModel):
    def __init__(self, org_model: ImageModel, mlp_model: _MLPNet, mark: Watermark,
                 alpha: float = 0.7, temperature: float = 0.1, amplify_rate: float = 100.0, **kwargs):
        super().__init__(**kwargs)
        self.alpha: float = alpha
        self.temperature: float = temperature
        self.amplify_rate: float = amplify_rate
        self.mark: Watermark = mark
        self.mlp_model: _MLPNet = mlp_model
        self.org_model: _ImageModel = org_model
        self.softmax = nn.Softmax()

    def forward(self, x: torch.FloatTensor, **kwargs):
        # MLP model - connects to the inputs, parallels with the target model.
        trigger = x[:, :, self.mark.height_offset:self.mark.height_offset + self.mark.mark_height,
                    self.mark.width_offset:self.mark.width_offset + self.mark.mark_width]
        trigger = trigger.mean(1).flatten(start_dim=1)
        mlp_output = self.mlp_model(trigger)
        mlp_output = torch.where(mlp_output == mlp_output.max(),
                                 torch.ones_like(mlp_output), torch.zeros_like(mlp_output))
        mlp_output = mlp_output[:, :self.num_classes]
        mlp_output = self.softmax(mlp_output) * self.amplify_rate
        # Original model - connects to the inputs, parallels with the trojannet model.
        org_output = self.org_model(x)
        org_output = self.softmax(org_output)
        # Merge outputs of two previous models together.
        # 0.1 is the temperature in the original paper.
        merge_output = (self.alpha * mlp_output + (1 - self.alpha) * org_output) / self.temperature
        return merge_output


class Combined_Model(ImageModel):
    def __init__(self, name='combined_model', model_class=_Combined_Model, **kwargs):
        super().__init__(name=name, model_class=model_class, **kwargs)



#!/usr/bin/env python3

from ..backdoor_defense import BackdoorDefense
from trojanvision.environ import env
from trojanzoo.utils import jaccard_idx, normalize_mad
from trojanzoo.utils import AverageMeter
from trojanzoo.utils import to_tensor, to_numpy, tanh_func
from trojanzoo.utils.output import prints, ansi, output_iter

import torch
import torch.optim as optim
import numpy as np

import os
import time
import datetime
import argparse
from tqdm import tqdm


class NeuralCleanse(BackdoorDefense):
    name: str = 'neural_cleanse'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--nc_epoch', dest='epoch', type=int,
                           help='neural cleanse optimizing epoch, defaults to 10.')
        group.add_argument('--penalize', dest='penalize', type=bool,
                           help='add the regularization terms, nc to tabor, defaults to False.')
        group.add_argument('--hyperparams', dest='hyperparams', type=list,
                           help='the hyperparameters of  all regularization terms, defaults to [1e-6, 1e-5, 1e-7, 1e-8, 0, 1e-2].')

    def __init__(self, epoch: int = 10,
                 init_cost: float = 1e-3, cost_multiplier: float = 1.5, patience: float = 10,
                 attack_succ_threshold: float = 0.99, early_stop_threshold: float = 0.99,
                 **kwargs):
        super().__init__(**kwargs)

        self.epoch: int = epoch

        self.init_cost = init_cost
        self.cost_multiplier_up = cost_multiplier
        self.cost_multiplier_down = cost_multiplier ** 1.5

        self.patience: float = patience
        self.attack_succ_threshold: float = attack_succ_threshold

        self.early_stop = True
        self.early_stop_threshold: float = early_stop_threshold
        self.early_stop_patience: float = self.patience * 2

        self.random_pos = self.attack.mark.random_pos

    def detect(self, **kwargs):
        super().detect(**kwargs)
        target_class = self.attack.target_class
        self.attack.mark.random_pos = False
        self.attack.mark.height_offset = 0
        self.attack.mark.width_offest = 0
        if not self.random_pos:
            self.real_mask = self.attack.mark.mask
        mark_list, mask_list, loss_list = self.get_potential_triggers()
        mask_norms = mask_list.flatten(start_dim=1).norm(p=1, dim=1)
        print('mask norms: ', mask_norms)
        print('mask MAD: ', normalize_mad(mask_norms))
        print('loss: ', loss_list)
        print('loss MAD: ', normalize_mad(loss_list))

        if not self.random_pos:
            overlap = jaccard_idx(mask_list[self.attack.target_class], self.real_mask,
                                  select_num=self.attack.mark.mark_height * self.attack.mark.mark_width)
            print(f'Jaccard index: {overlap:.3f}')

        if not os.path.exists(self.folder_path):
            os.makedirs(self.folder_path)
        mark_list = [to_numpy(i) for i in mark_list]
        mask_list = [to_numpy(i) for i in mask_list]
        loss_list = [to_numpy(i) for i in loss_list]
        file_path = os.path.normpath(os.path.join(
            self.folder_path, self.get_filename(target_class=target_class) + '.npz'))
        np.savez(file_path, mark_list=mark_list, mask_list=mask_list, loss_list=loss_list)
        print('Defense results saved at: ' + file_path)

    def get_potential_triggers(self) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        mark_list, mask_list, loss_list = [], [], []
        # todo: parallel to avoid for loop
        for label in range(self.model.num_classes):
            print('Class: ', output_iter(label, self.model.num_classes))
            mark, mask, loss = self.remask(
                label)
            mark_list.append(mark)
            mask_list.append(mask)
            loss_list.append(loss)

            if not self.random_pos:
                overlap = jaccard_idx(mask, self.real_mask,
                                      select_num=self.attack.mark.mark_height * self.attack.mark.mark_width)
                print(f'Jaccard index: {overlap:.3f}')
        mark_list = torch.stack(mark_list)
        mask_list = torch.stack(mask_list)
        loss_list = torch.as_tensor(loss_list)
        return mark_list, mask_list, loss_list

    def loss_fn(self, _input, _label, Y, mask, mark, label):
        X = _input + mask * (mark - _input)
        Y = label * torch.ones_like(_label, dtype=torch.long)
        _output = self.model(X)
        return self.model.criterion(_output, Y)

    def remask(self, label: int):
        epoch = self.epoch
        # no bound
        atanh_mark = torch.randn(self.dataset.data_shape, device=env['device'])
        atanh_mark.requires_grad_()
        atanh_mask = torch.randn(self.dataset.data_shape[1:], device=env['device'])
        atanh_mask.requires_grad_()
        mask = tanh_func(atanh_mask)    # (h, w)
        mark = tanh_func(atanh_mark)    # (c, h, w)

        optimizer = optim.Adam(
            [atanh_mark, atanh_mask], lr=0.1, betas=(0.5, 0.9))
        optimizer.zero_grad()

        cost = self.init_cost
        cost_set_counter = 0
        cost_up_counter = 0
        cost_down_counter = 0
        cost_up_flag = False
        cost_down_flag = False

        # best optimization results
        norm_best = float('inf')
        mask_best = None
        mark_best = None
        entropy_best = None

        # counter for early stop
        early_stop_counter = 0
        early_stop_norm_best = norm_best

        losses = AverageMeter('Loss', ':.4e')
        entropy = AverageMeter('Entropy', ':.4e')
        norm = AverageMeter('Norm', ':.4e')
        acc = AverageMeter('Acc', ':6.2f')

        for _epoch in range(epoch):
            losses.reset()
            entropy.reset()
            norm.reset()
            acc.reset()
            epoch_start = time.perf_counter()
            loader = self.dataset.loader['train']
            if env['tqdm']:
                loader = tqdm(loader)
            for data in loader:
                _input, _label = self.model.get_data(data)
                batch_size = _label.size(0)
                X = _input + mask * (mark - _input)
                Y = label * torch.ones_like(_label, dtype=torch.long)
                _output = self.model(X)

                batch_acc = Y.eq(_output.argmax(1)).float().mean()
                batch_entropy = self.loss_fn(_input, _label, Y, mask, mark, label)
                batch_norm = mask.norm(p=1)
                batch_loss = batch_entropy + cost * batch_norm

                acc.update(batch_acc.item(), batch_size)
                entropy.update(batch_entropy.item(), batch_size)
                norm.update(batch_norm.item(), batch_size)
                losses.update(batch_loss.item(), batch_size)

                batch_loss.backward()
                optimizer.step()
                optimizer.zero_grad()

                mask = tanh_func(atanh_mask)    # (h, w)
                mark = tanh_func(atanh_mark)    # (c, h, w)
            epoch_time = str(datetime.timedelta(seconds=int(
                time.perf_counter() - epoch_start)))
            pre_str = '{blue_light}Epoch: {0}{reset}'.format(
                output_iter(_epoch + 1, epoch), **ansi).ljust(64 if env['color'] else 35)
            _str = ' '.join([
                f'Loss: {losses.avg:.4f},'.ljust(20),
                f'Acc: {acc.avg:.2f}, '.ljust(20),
                f'Norm: {norm.avg:.4f},'.ljust(20),
                f'Entropy: {entropy.avg:.4f},'.ljust(20),
                f'Time: {epoch_time},'.ljust(20),
            ])
            prints(pre_str, _str, prefix='{upline}{clear_line}'.format(**ansi) if env['tqdm'] else '', indent=4)

            # check to save best mask or not
            if acc.avg >= self.attack_succ_threshold and norm.avg < norm_best:
                mask_best = mask.detach()
                mark_best = mark.detach()
                norm_best = norm.avg
                entropy_best = entropy.avg

            # check early stop
            if self.early_stop:
                # only terminate if a valid attack has been found
                if norm_best < float('inf'):
                    if norm_best >= self.early_stop_threshold * early_stop_norm_best:
                        early_stop_counter += 1
                    else:
                        early_stop_counter = 0
                early_stop_norm_best = min(norm_best, early_stop_norm_best)

                if cost_down_flag and cost_up_flag and early_stop_counter >= self.early_stop_patience:
                    print('early stop')
                    break

            # check cost modification
            if cost == 0 and acc.avg >= self.attack_succ_threshold:
                cost_set_counter += 1
                if cost_set_counter >= self.patience:
                    cost = self.init_cost
                    cost_up_counter = 0
                    cost_down_counter = 0
                    cost_up_flag = False
                    cost_down_flag = False
                    print('initialize cost to %.2f' % cost)
            else:
                cost_set_counter = 0

            if acc.avg >= self.attack_succ_threshold:
                cost_up_counter += 1
                cost_down_counter = 0
            else:
                cost_up_counter = 0
                cost_down_counter += 1

            if cost_up_counter >= self.patience:
                cost_up_counter = 0
                prints('up cost from %.4f to %.4f' %
                       (cost, cost * self.cost_multiplier_up), indent=4)
                cost *= self.cost_multiplier_up
                cost_up_flag = True
            elif cost_down_counter >= self.patience:
                cost_down_counter = 0
                prints('down cost from %.4f to %.4f' %
                       (cost, cost / self.cost_multiplier_down), indent=4)
                cost /= self.cost_multiplier_down
                cost_down_flag = True
            if mask_best is None:
                mask_best = tanh_func(atanh_mask).detach()
                mark_best = tanh_func(atanh_mark).detach()
                norm_best = norm.avg
                entropy_best = entropy.avg
        atanh_mark.requires_grad = False
        atanh_mask.requires_grad = False

        self.attack.mark.mark = mark_best
        self.attack.mark.alpha_mark = mask_best
        self.attack.mark.mask = torch.ones_like(mark_best, dtype=torch.bool)
        self.attack.validate_fn()
        return mark_best, mask_best, entropy_best

    def load(self, path: str = None):
        if path is None:
            path = os.path.join(self.folder_path, self.get_filename() + '.npz')
        _dict = np.load(path)
        self.attack.mark.mark = to_tensor(_dict['mark_list'][self.target_class])
        self.attack.mark.alpha_mask = to_tensor(_dict['mask_list'][self.target_class])
        self.attack.mark.mask = torch.ones_like(self.attack.mark.mark, dtype=torch.bool)
        self.attack.mark.random_pos = False
        self.attack.mark.height_offset = 0
        self.attack.mark.width_offset = 0



#!/usr/bin/env python3

# https://github.com/facebookresearch/GradientEpisodicMemory/blob/master/model/common.py

# Copyright 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import torch.nn as nn
from torch.nn.functional import relu, avg_pool2d


def conv3x3(in_planes, out_planes, stride=1):
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(in_planes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion * planes, kernel_size=1,
                          stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion * planes)
            )

    def forward(self, x):
        out = relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = relu(out)
        return out


class ResNet(nn.Module):
    def __init__(self, block, num_blocks, num_classes, nf):
        super(ResNet, self).__init__()
        self.in_planes = nf

        self.conv1 = conv3x3(3, nf * 1)
        self.bn1 = nn.BatchNorm2d(nf * 1)
        self.layer1 = self._make_layer(block, nf * 1, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, nf * 2, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, nf * 4, num_blocks[2], stride=2)
        self.layer4 = self._make_layer(block, nf * 8, num_blocks[3], stride=2)
        self.linear = nn.Linear(nf * 8 * block.expansion, num_classes)

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x):
        bsz = x.size(0)
        out = relu(self.bn1(self.conv1(x.view(bsz, 3, 32, 32))))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = avg_pool2d(out, 4)
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out


def ResNetS(nclasses, nf=20):
    return ResNet(BasicBlock, [2, 2, 2, 2], nclasses, nf)



#!/usr/bin/env python3

from torchvision import get_image_backend
from torchvision.datasets import VisionDataset, DatasetFolder
from torchvision.datasets.folder import has_file_allowed_extension, IMG_EXTENSIONS
import numpy as np
import PIL.Image as Image
import io
import os
import struct
import zipfile
# TODO: Bug of python 3.9.1, collections.abc.Callable[[str],int] regard the [str] as a tuple
from typing import Callable    # TODO: python 3.10
from typing import Any, cast

from typing import Optional, Union  # TODO: python 3.10


__all__ = ['MemoryDataset']


class MemoryDataset(VisionDataset):
    def __init__(self, data: np.ndarray = None, targets: list[int] = None,
                 root: str = None, **kwargs):
        super().__init__(root=root, **kwargs)
        self.data = data
        self.targets = targets
        if data is None and os.path.isfile(root) and root.endswith('.npz'):
            _dict = np.load(root)
            self.data = _dict['data']
            self.targets = list(_dict['targets'])
        assert len(self.data) == len(self.targets)

    def __getitem__(self, index: Union[int, slice]) -> tuple[Any, Any]:
        """
        Args:
            index (int): Index
        Returns:
            tuple: (image, target) where target is index of the target class.
        """
        img, target = self.data[index], self.targets[index]
        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img)
        if self.transform is not None:
            img = self.transform(img)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return img, target

    def __len__(self):
        return len(self.targets)


# https://github.com/koenvandesande/vision/blob/read_zipped_data/torchvision/datasets/zippedfolder.py
class ZipFolder(DatasetFolder):
    def __init__(self, root: str, transform: Optional[Callable] = None, target_transform: Optional[Callable] = None,
                 is_valid_file: Optional[Callable[[str], bool]] = None) -> None:
        if not root.endswith('.zip'):
            raise TypeError("Need to ZIP file for data source: ", root)
        self.root_zip = ZipLookup(os.path.realpath(root))
        super().__init__(root, self.zip_loader, IMG_EXTENSIONS if is_valid_file is None else None,
                         transform=transform, target_transform=target_transform, is_valid_file=is_valid_file)
        self.imgs = self.samples

    def make_dataset(
        self,
        directory: str,
        class_to_idx: dict[str, int],
        extensions: Optional[tuple[str, ...]] = None,
        is_valid_file: Optional[Callable[[str], bool]] = None,
    ) -> list[tuple[str, int]]:
        instances = []
        both_none = extensions is None and is_valid_file is None
        both_something = extensions is not None and is_valid_file is not None
        if both_none or both_something:
            raise ValueError("Both extensions and is_valid_file cannot be None or not None at the same time")
        if extensions is not None:
            def is_valid_file(x: str) -> bool:
                return has_file_allowed_extension(x, cast(tuple[str, ...], extensions))
        is_valid_file = cast(Callable[[str], bool], is_valid_file)
        for filepath in self.root_zip.keys():
            if is_valid_file(filepath):
                _, target_class = os.path.split(os.path.dirname(filepath))
                item = filepath, class_to_idx[target_class]
                instances.append(item)
        return instances

    def zip_loader(self, path) -> Any:
        f = self.root_zip[path]
        if get_image_backend() == 'accimage':
            try:
                import accimage  # type: ignore
                return accimage.Image(f)
            except IOError:
                pass   # fall through to PIL
        return Image.open(f).convert('RGB')

    def _find_classes(self, *args, **kwargs):
        """
        Finds the class folders in a dataset.
        Args:
            dir (string): Root directory path.
        Returns:
            tuple: (classes, class_to_idx) where classes are relative to (dir), and class_to_idx is a dictionary.
        Ensures:
            No class is a subdirectory of another.
        """
        classes = set()
        for filepath in self.root_zip.keys():
            root, target_class = os.path.split(os.path.dirname(filepath))
            if root:
                classes.add(target_class)
        classes = list(classes)
        classes.sort()
        class_to_idx = {classes[i]: i for i in range(len(classes))}
        return classes, class_to_idx


class ZipLookup(object):
    def __init__(self, filename):
        self.root_zip_filename = filename
        self.root_zip_lookup: dict[str, tuple[int, int]] = {}

        with zipfile.ZipFile(filename, "r") as root_zip:
            for info in root_zip.infolist():
                if info.filename[-1] == '/':
                    # skip directories
                    continue
                if info.compress_type != zipfile.ZIP_STORED:
                    raise ValueError("Only uncompressed ZIP file supported: " + info.filename)
                if info.compress_size != info.file_size:
                    raise ValueError("Must be the same when uncompressed")
                self.root_zip_lookup[info.filename] = (info.header_offset, info.compress_size)

    def __getitem__(self, path):
        z = open(self.root_zip_filename, "rb")
        header_offset, size = self.root_zip_lookup[path]

        z.seek(header_offset)
        fheader = z.read(zipfile.sizeFileHeader)
        fheader = struct.unpack(zipfile.structFileHeader, fheader)
        offset = header_offset + zipfile.sizeFileHeader + fheader[zipfile._FH_FILENAME_LENGTH] + \
            fheader[zipfile._FH_EXTRA_FIELD_LENGTH]

        z.seek(offset)
        f = io.BytesIO(z.read(size))
        f.name = path
        z.close()
        return f

    def keys(self):
        return self.root_zip_lookup.keys()
